#!/usr/bin/python3
import os
import shutil
from errno import ENOENT
from subprocess import Popen, PIPE
from argparse import ArgumentParser


DEFAULT_GLUSTERD_WORKDIR = "/var/lib/glusterd"


def handle_rm_error(func, path, exc_info):
    if exc_info[1].errno == ENOENT:
        return

    raise exc_info[1]


def get_glusterd_workdir():
    p = Popen(["gluster", "system::", "getwd"],
              stdout=PIPE, stderr=PIPE, universal_newlines=True)

    out, _ = p.communicate()

    if p.returncode == 0:
        return out.strip()
    else:
        return DEFAULT_GLUSTERD_WORKDIR


def get_args():
    parser = ArgumentParser(description="Volume delete post hook script")
    parser.add_argument("--volname")
    return parser.parse_args()


def main():
    args = get_args()
    glusterfind_dir = os.path.join(get_glusterd_workdir(), "glusterfind")

    # Check all session directories, if any directory found for
    # the deleted volume, cleanup all the session directories
    try:
        ls_glusterfind_dir = os.listdir(glusterfind_dir)
    except OSError:
        ls_glusterfind_dir = []

    for session in ls_glusterfind_dir:
        # don't blow away the keys directory
        if session == ".keys":
            continue

        # Possible session directory
        volume_session_path = os.path.join(glusterfind_dir,
                                           session,
                                           args.volname)
        if os.path.exists(volume_session_path):
            shutil.rmtree(volume_session_path, onerror=handle_rm_error)

        # Try to Remove directory, if any other dir exists for different
        # volume, then rmdir will fail with ENOTEMPTY which is fine
        try:
            os.rmdir(os.path.join(glusterfind_dir, session))
        except (OSError, IOError):
            pass


if __name__ == "__main__":
    main()



#!/usr/bin/python3
# -*- coding: utf-8 -*-

# Copyright (c) 2015 Red Hat, Inc. <http://www.redhat.com/>
# This file is part of GlusterFS.
#
# This file is licensed to you under your choice of the GNU Lesser
# General Public License, version 3 or any later version (LGPLv3 or
# later), or the GNU General Public License, version 2 (GPLv2), in all
# cases as published by the Free Software Foundation.

import os
import sys
import logging
from argparse import ArgumentParser, RawDescriptionHelpFormatter
try:
    import urllib.parse as urllib
except ImportError:
    import urllib
import time

from utils import mkdirp, setup_logger, create_file, output_write, find
import conf


PROG_DESCRIPTION = """
Changelog Crawler
"""

logger = logging.getLogger()


def brickfind_crawl(brick, args):
    if brick.endswith("/"):
        brick = brick[0:len(brick)-1]

    working_dir = os.path.dirname(args.outfile)
    mkdirp(working_dir, exit_on_err=True, logger=logger)
    create_file(args.outfile, exit_on_err=True, logger=logger)

    with open(args.outfile, "a+") as fout:
        brick_path_len = len(brick)

        def output_callback(path, filter_result, is_dir):
            path = path.strip()
            path = path[brick_path_len+1:]

            if args.type == "both":
                output_write(fout, path, args.output_prefix,
                             encode=(not args.no_encode), tag=args.tag,
                             field_separator=args.field_separator)
            else:
                if (is_dir and args.type == "d") or (
                    (not is_dir) and args.type == "f"):
                    output_write(fout, path, args.output_prefix,
                    encode=(not args.no_encode), tag=args.tag,
                    field_separator=args.field_separator)

        ignore_dirs = [os.path.join(brick, dirname)
                       for dirname in
                       conf.get_opt("brick_ignore_dirs").split(",")]

        find(brick, callback_func=output_callback,
             ignore_dirs=ignore_dirs)

        fout.flush()
        os.fsync(fout.fileno())


def _get_args():
    parser = ArgumentParser(formatter_class=RawDescriptionHelpFormatter,
                            description=PROG_DESCRIPTION)

    parser.add_argument("session", help="Session Name")
    parser.add_argument("volume", help="Volume Name")
    parser.add_argument("node", help="Node Name")
    parser.add_argument("brick", help="Brick Name")
    parser.add_argument("outfile", help="Output File")
    parser.add_argument("tag", help="Tag to prefix file name with")
    parser.add_argument("--only-query", help="Only query, No session update",
                        action="store_true")
    parser.add_argument("--debug", help="Debug", action="store_true")
    parser.add_argument("--no-encode",
                        help="Do not encode path in outfile",
                        action="store_true")
    parser.add_argument("--output-prefix", help="File prefix in output",
                        default=".")
    parser.add_argument('--type', help="type: f, f-files only"
                        " d, d-directories only, by default = both",
                        default='both')
    parser.add_argument("--field-separator", help="Field separator",
                        default=" ")

    return parser.parse_args()


if __name__ == "__main__":
    args = _get_args()
    session_dir = os.path.join(conf.get_opt("session_dir"), args.session)
    status_file = os.path.join(session_dir, args.volume,
                     "%s.status" % urllib.quote_plus(args.brick))
    status_file_pre = status_file + ".pre"
    mkdirp(os.path.join(session_dir, args.volume), exit_on_err=True,
           logger=logger)
    mkdirp(os.path.join(conf.get_opt("log_dir"), args.session, args.volume),
           exit_on_err=True)
    log_file = os.path.join(conf.get_opt("log_dir"),
                            args.session,
                            args.volume,
                            "brickfind.log")
    setup_logger(logger, log_file, args.debug)

    time_to_update = int(time.time())
    brickfind_crawl(args.brick, args)
    if not args.only_query:
        with open(status_file_pre, "w") as f:
            f.write(str(time_to_update))
    sys.exit(0)



# -*- coding: utf-8 -*-
#
#  Copyright (c) 2016 Red Hat, Inc. <http://www.redhat.com>
#  This file is part of GlusterFS.
#
#  This file is licensed to you under your choice of the GNU Lesser
#  General Public License, version 3 or any later version (LGPLv3 or
#  later), or the GNU General Public License, version 2 (GPLv2), in all
#  cases as published by the Free Software Foundation.
#

import subprocess
glusterd_workdir = None

# Methods
def get_glusterd_workdir():
    global glusterd_workdir
    if glusterd_workdir is not None:
        return glusterd_workdir
    proc = subprocess.Popen(["gluster", "system::", "getwd"],
                            stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                            universal_newlines = True)
    out, err = proc.communicate()
    if proc.returncode == 0:
        glusterd_workdir = out.strip()
    else:
        glusterd_workdir = "/var/lib/glusterd"
    return glusterd_workdir

SERVER_ADDRESS = "0.0.0.0"
SERVER_ADDRESSv4 = "0.0.0.0"
SERVER_ADDRESSv6 = "::1"
DEFAULT_CONFIG_FILE = "/etc/glusterfs/eventsconfig.json"
CUSTOM_CONFIG_FILE_TO_SYNC = "/events/config.json"
CUSTOM_CONFIG_FILE = get_glusterd_workdir() + CUSTOM_CONFIG_FILE_TO_SYNC
WEBHOOKS_FILE_TO_SYNC = "/events/webhooks.json"
WEBHOOKS_FILE = get_glusterd_workdir() + WEBHOOKS_FILE_TO_SYNC
LOG_FILE = "/var/log/glusterfs/events.log"
EVENTSD = "glustereventsd"
CONFIG_KEYS = ["log-level", "port", "disable-events-log"]
BOOL_CONFIGS = ["disable-events-log"]
INT_CONFIGS = ["port"]
RESTART_CONFIGS = ["port"]
EVENTS_ENABLED = 1
UUID_FILE = get_glusterd_workdir() + "/glusterd.info"
PID_FILE = "/var/run/glustereventsd.pid"
AUTO_BOOL_ATTRIBUTES = ["force", "push-pem", "no-verify"]
AUTO_INT_ATTRIBUTES = ["ssh-port"]
CERTS_DIR = get_glusterd_workdir() + "/events"

# Errors
ERROR_SAME_CONFIG = 2
ERROR_ALL_NODES_STATUS_NOT_OK = 3
ERROR_PARTIAL_SUCCESS = 4
ERROR_WEBHOOK_ALREADY_EXISTS = 5
ERROR_WEBHOOK_NOT_EXISTS = 6
ERROR_INVALID_CONFIG = 7
ERROR_WEBHOOK_SYNC_FAILED = 8
ERROR_CONFIG_SYNC_FAILED = 9



#
# Copyright (c) 2018 Red Hat, Inc. <http://www.redhat.com>
# This file is part of GlusterFS.

# This file is licensed to you under your choice of the GNU Lesser
# General Public License, version 3 or any later version (LGPLv3 or
# later), or the GNU General Public License, version 2 (GPLv2), in all
# cases as published by the Free Software Foundation.
#

# All python2/python3 compatibility routines

import sys
import os
import stat
import struct
from syncdutils import umask
from ctypes import create_string_buffer

if sys.version_info >= (3,):
    def pipe():
        (r, w) = os.pipe()
        os.set_inheritable(r, True)
        os.set_inheritable(w, True)
        return (r, w)

    # Raw conversion of bytearray to string. Used in the cases where
    # buffer is created by create_string_buffer which is a 8-bit char
    # array and passed to syscalls to fetch results. Using encode/decode
    # doesn't work as it converts to string altering the size.
    def bytearray_to_str(byte_arr):
        return ''.join([chr(b) for b in byte_arr])

    # Raw conversion of string to bytes. This is required to convert
    # back the string into bytearray(c char array) to use in struc
    # pack/unpacking. Again encode/decode can't be used as it
    # converts it alters size.
    def str_to_bytearray(string):
        return bytes([ord(c) for c in string])

    def gr_create_string_buffer(size):
        return create_string_buffer(b'\0', size)

    def gr_query_xattr(cls, path, size, syscall, attr=None):
        if attr:
            return cls._query_xattr(path.encode(), size, syscall,
                                    attr.encode())
        else:
            return cls._query_xattr(path.encode(), size, syscall)

    def gr_lsetxattr(cls, path, attr, val):
        return cls.libc.lsetxattr(path.encode(), attr.encode(), val,
                                  len(val), 0)

    def gr_lremovexattr(cls, path, attr):
        return cls.libc.lremovexattr(path.encode(), attr.encode())

    def gr_cl_register(libgfapi, brick, path, log_file, log_level, retries):
        return libgfapi.gf_changelog_register(brick.encode(),
                                              path.encode(),
                                              log_file.encode(),
                                              log_level, retries)

    def gr_cl_done(libgfapi, clfile):
        return libgfapi.gf_changelog_done(clfile.encode())

    def gr_cl_history_changelog(libgfapi, changelog_path, start, end, num_parallel,
                                actual_end):
        return libgfapi.gf_history_changelog(changelog_path.encode(),
                                             start, end, num_parallel,
                                             actual_end)

    def gr_cl_history_done(libgfapi, clfile):
        return libgfapi.gf_history_changelog_done(clfile.encode())

    # regular file

    def entry_pack_reg(cls, gf, bn, mo, uid, gid):
        bn_encoded = bn.encode()
        blen = len(bn_encoded)
        return struct.pack(cls._fmt_mknod(blen),
                           uid, gid, gf.encode(), mo, bn_encoded,
                           stat.S_IMODE(mo), 0, umask())

    def entry_pack_reg_stat(cls, gf, bn, st):
        bn_encoded = bn.encode()
        blen = len(bn_encoded)
        mo = st['mode']
        return struct.pack(cls._fmt_mknod(blen),
                           st['uid'], st['gid'],
                           gf.encode(), mo, bn_encoded,
                           stat.S_IMODE(mo), 0, umask())
    # mkdir

    def entry_pack_mkdir(cls, gf, bn, mo, uid, gid):
        bn_encoded = bn.encode()
        blen = len(bn_encoded)
        return struct.pack(cls._fmt_mkdir(blen),
                           uid, gid, gf.encode(), mo, bn_encoded,
                           stat.S_IMODE(mo), umask())
    # symlink

    def entry_pack_symlink(cls, gf, bn, lnk, st):
        bn_encoded = bn.encode()
        blen = len(bn_encoded)
        lnk_encoded = lnk.encode()
        llen = len(lnk_encoded)
        return struct.pack(cls._fmt_symlink(blen, llen),
                           st['uid'], st['gid'],
                           gf.encode(), st['mode'], bn_encoded,
                           lnk_encoded)
else:
    def pipe():
        (r, w) = os.pipe()
        return (r, w)

    # Raw conversion of bytearray to string
    def bytearray_to_str(byte_arr):
        return byte_arr

    # Raw conversion of string to bytearray
    def str_to_bytearray(string):
        return string

    def gr_create_string_buffer(size):
        return create_string_buffer('\0', size)

    def gr_query_xattr(cls, path, size, syscall, attr=None):
        if attr:
            return cls._query_xattr(path, size, syscall, attr)
        else:
            return cls._query_xattr(path, size, syscall)

    def gr_lsetxattr(cls, path, attr, val):
        return cls.libc.lsetxattr(path, attr, val, len(val), 0)

    def gr_lremovexattr(cls, path, attr):
        return cls.libc.lremovexattr(path, attr)

    def gr_cl_register(libgfapi, brick, path, log_file, log_level, retries):
        return libgfapi.gf_changelog_register(brick, path, log_file,
                                              log_level, retries)

    def gr_cl_done(libgfapi, clfile):
        return libgfapi.gf_changelog_done(clfile)

    def gr_cl_history_changelog(libgfapi, changelog_path, start, end, num_parallel,
                                actual_end):
        return libgfapi.gf_history_changelog(changelog_path, start, end,
                                             num_parallel, actual_end)

    def gr_cl_history_done(libgfapi, clfile):
        return libgfapi.gf_history_changelog_done(clfile)

    # regular file

    def entry_pack_reg(cls, gf, bn, mo, uid, gid):
        blen = len(bn)
        return struct.pack(cls._fmt_mknod(blen),
                           uid, gid, gf, mo, bn,
                           stat.S_IMODE(mo), 0, umask())

    def entry_pack_reg_stat(cls, gf, bn, st):
        blen = len(bn)
        mo = st['mode']
        return struct.pack(cls._fmt_mknod(blen),
                           st['uid'], st['gid'],
                           gf, mo, bn,
                           stat.S_IMODE(mo), 0, umask())
    # mkdir

    def entry_pack_mkdir(cls, gf, bn, mo, uid, gid):
        blen = len(bn)
        return struct.pack(cls._fmt_mkdir(blen),
                           uid, gid, gf, mo, bn,
                           stat.S_IMODE(mo), umask())
    # symlink

    def entry_pack_symlink(cls, gf, bn, lnk, st):
        blen = len(bn)
        llen = len(lnk)
        return struct.pack(cls._fmt_symlink(blen, llen),
                           st['uid'], st['gid'],
                           gf, st['mode'], bn, lnk)



# -*- coding: utf-8 -*-
all_events = [
    "EVENT_PEER_ATTACH",
    "EVENT_PEER_DETACH",
    "EVENT_VOLUME_CREATE",
    "EVENT_VOLUME_START",
    "EVENT_VOLUME_STOP",
    "EVENT_VOLUME_DELETE",
    "EVENT_VOLUME_SET",
    "EVENT_VOLUME_RESET",
    "EVENT_BRICK_RESET_START",
    "EVENT_BRICK_RESET_COMMIT",
    "EVENT_BRICK_REPLACE",
    "EVENT_GEOREP_CREATE",
    "EVENT_GEOREP_START",
    "EVENT_GEOREP_STOP",
    "EVENT_GEOREP_PAUSE",
    "EVENT_GEOREP_RESUME",
    "EVENT_GEOREP_DELETE",
    "EVENT_GEOREP_CONFIG_SET",
    "EVENT_GEOREP_CONFIG_RESET",
    "EVENT_BITROT_ENABLE",
    "EVENT_BITROT_DISABLE",
    "EVENT_BITROT_SCRUB_THROTTLE",
    "EVENT_BITROT_SCRUB_FREQ",
    "EVENT_BITROT_SCRUB_OPTION",
    "EVENT_BITROT_SCRUB_ONDEMAND",
    "EVENT_QUOTA_ENABLE",
    "EVENT_QUOTA_DISABLE",
    "EVENT_QUOTA_SET_USAGE_LIMIT",
    "EVENT_QUOTA_SET_OBJECTS_LIMIT",
    "EVENT_QUOTA_REMOVE_USAGE_LIMIT",
    "EVENT_QUOTA_REMOVE_OBJECTS_LIMIT",
    "EVENT_QUOTA_ALERT_TIME",
    "EVENT_QUOTA_SOFT_TIMEOUT",
    "EVENT_QUOTA_HARD_TIMEOUT",
    "EVENT_QUOTA_DEFAULT_SOFT_LIMIT",
    "EVENT_SNAPSHOT_CREATED",
    "EVENT_SNAPSHOT_CREATE_FAILED",
    "EVENT_SNAPSHOT_ACTIVATED",
    "EVENT_SNAPSHOT_ACTIVATE_FAILED",
    "EVENT_SNAPSHOT_DEACTIVATED",
    "EVENT_SNAPSHOT_DEACTIVATE_FAILED",
    "EVENT_SNAPSHOT_SOFT_LIMIT_REACHED",
    "EVENT_SNAPSHOT_HARD_LIMIT_REACHED",
    "EVENT_SNAPSHOT_RESTORED",
    "EVENT_SNAPSHOT_RESTORE_FAILED",
    "EVENT_SNAPSHOT_DELETED",
    "EVENT_SNAPSHOT_DELETE_FAILED",
    "EVENT_SNAPSHOT_CLONED",
    "EVENT_SNAPSHOT_CLONE_FAILED",
    "EVENT_SNAPSHOT_CONFIG_UPDATED",
    "EVENT_SNAPSHOT_CONFIG_UPDATE_FAILED",
    "EVENT_SNAPSHOT_SCHEDULER_INITIALISED",
    "EVENT_SNAPSHOT_SCHEDULER_INIT_FAILED",
    "EVENT_SNAPSHOT_SCHEDULER_ENABLED",
    "EVENT_SNAPSHOT_SCHEDULER_ENABLE_FAILED",
    "EVENT_SNAPSHOT_SCHEDULER_DISABLED",
    "EVENT_SNAPSHOT_SCHEDULER_DISABLE_FAILED",
    "EVENT_SNAPSHOT_SCHEDULER_SCHEDULE_ADDED",
    "EVENT_SNAPSHOT_SCHEDULER_SCHEDULE_ADD_FAILED",
    "EVENT_SNAPSHOT_SCHEDULER_SCHEDULE_EDITED",
    "EVENT_SNAPSHOT_SCHEDULER_SCHEDULE_EDIT_FAILED",
    "EVENT_SNAPSHOT_SCHEDULER_SCHEDULE_DELETED",
    "EVENT_SNAPSHOT_SCHEDULER_SCHEDULE_DELETE_FAILED",
    "EVENT_SVC_MANAGER_FAILED",
    "EVENT_SVC_RECONFIGURE_FAILED",
    "EVENT_SVC_CONNECTED",
    "EVENT_SVC_DISCONNECTED",
    "EVENT_PEER_STORE_FAILURE",
    "EVENT_PEER_RPC_CREATE_FAILED",
    "EVENT_PEER_REJECT",
    "EVENT_PEER_CONNECT",
    "EVENT_PEER_DISCONNECT",
    "EVENT_PEER_NOT_FOUND",
    "EVENT_UNKNOWN_PEER",
    "EVENT_BRICK_START_FAILED",
    "EVENT_BRICK_STOP_FAILED",
    "EVENT_BRICK_DISCONNECTED",
    "EVENT_BRICK_CONNECTED",
    "EVENT_BRICKPATH_RESOLVE_FAILED",
    "EVENT_NOTIFY_UNKNOWN_OP",
    "EVENT_QUORUM_LOST",
    "EVENT_QUORUM_REGAINED",
    "EVENT_REBALANCE_START_FAILED",
    "EVENT_REBALANCE_STATUS_UPDATE_FAILED",
    "EVENT_IMPORT_QUOTA_CONF_FAILED",
    "EVENT_IMPORT_VOLUME_FAILED",
    "EVENT_IMPORT_BRICK_FAILED",
    "EVENT_COMPARE_FRIEND_VOLUME_FAILED",
    "EVENT_NFS_GANESHA_EXPORT_FAILED",
    "EVENT_EC_MIN_BRICKS_NOT_UP",
    "EVENT_EC_MIN_BRICKS_UP",
    "EVENT_GEOREP_FAULTY",
    "EVENT_GEOREP_CHECKPOINT_COMPLETED",
    "EVENT_GEOREP_ACTIVE",
    "EVENT_GEOREP_PASSIVE",
    "EVENT_QUOTA_CROSSED_SOFT_LIMIT",
    "EVENT_BITROT_BAD_FILE",
    "EVENT_CLIENT_CONNECT",
    "EVENT_CLIENT_AUTH_REJECT",
    "EVENT_CLIENT_DISCONNECT",
    "EVENT_POSIX_SAME_GFID",
    "EVENT_POSIX_ALREADY_PART_OF_VOLUME",
    "EVENT_POSIX_BRICK_NOT_IN_VOLUME",
    "EVENT_POSIX_BRICK_VERIFICATION_FAILED",
    "EVENT_POSIX_ACL_NOT_SUPPORTED",
    "EVENT_POSIX_HEALTH_CHECK_FAILED",
    "EVENT_AFR_QUORUM_MET",
    "EVENT_AFR_QUORUM_FAIL",
    "EVENT_AFR_SUBVOL_UP",
    "EVENT_AFR_SUBVOLS_DOWN",
    "EVENT_AFR_SPLIT_BRAIN",
    "EVENT_TIER_ATTACH",
    "EVENT_TIER_ATTACH_FORCE",
    "EVENT_TIER_DETACH_START",
    "EVENT_TIER_DETACH_STOP",
    "EVENT_TIER_DETACH_COMMIT",
    "EVENT_TIER_DETACH_FORCE",
    "EVENT_TIER_PAUSE",
    "EVENT_TIER_RESUME",
    "EVENT_TIER_WATERMARK_HI",
    "EVENT_TIER_WATERMARK_DROPPED_TO_MID",
    "EVENT_TIER_WATERMARK_RAISED_TO_MID",
    "EVENT_TIER_WATERMARK_DROPPED_TO_LOW",
    "EVENT_VOLUME_ADD_BRICK",
    "EVENT_VOLUME_ADD_BRICK_FAILED",
    "EVENT_VOLUME_REMOVE_BRICK_START",
    "EVENT_VOLUME_REMOVE_BRICK_START_FAILED",
    "EVENT_VOLUME_REMOVE_BRICK_COMMIT",
    "EVENT_VOLUME_REMOVE_BRICK_COMMIT_FAILED",
    "EVENT_VOLUME_REMOVE_BRICK_STOP",
    "EVENT_VOLUME_REMOVE_BRICK_STOP_FAILED",
    "EVENT_VOLUME_REMOVE_BRICK_FORCE",
    "EVENT_VOLUME_REMOVE_BRICK_FORCE_FAILED",
    "EVENT_VOLUME_REMOVE_BRICK_FAILED",
    "EVENT_VOLUME_REBALANCE_START",
    "EVENT_VOLUME_REBALANCE_STOP",
    "EVENT_VOLUME_REBALANCE_FAILED",
    "EVENT_VOLUME_REBALANCE_COMPLETE",
    "EVENT_TIER_START",
    "EVENT_TIER_START_FORCE",
    "EVENT_DHT_DISK_USAGE",
    "EVENT_DHT_INODES_USAGE",
]

PEER_ATTACH = 0
PEER_DETACH = 1
VOLUME_CREATE = 2
VOLUME_START = 3
VOLUME_STOP = 4
VOLUME_DELETE = 5
VOLUME_SET = 6
VOLUME_RESET = 7
BRICK_RESET_START = 8
BRICK_RESET_COMMIT = 9
BRICK_REPLACE = 10
GEOREP_CREATE = 11
GEOREP_START = 12
GEOREP_STOP = 13
GEOREP_PAUSE = 14
GEOREP_RESUME = 15
GEOREP_DELETE = 16
GEOREP_CONFIG_SET = 17
GEOREP_CONFIG_RESET = 18
BITROT_ENABLE = 19
BITROT_DISABLE = 20
BITROT_SCRUB_THROTTLE = 21
BITROT_SCRUB_FREQ = 22
BITROT_SCRUB_OPTION = 23
BITROT_SCRUB_ONDEMAND = 24
QUOTA_ENABLE = 25
QUOTA_DISABLE = 26
QUOTA_SET_USAGE_LIMIT = 27
QUOTA_SET_OBJECTS_LIMIT = 28
QUOTA_REMOVE_USAGE_LIMIT = 29
QUOTA_REMOVE_OBJECTS_LIMIT = 30
QUOTA_ALERT_TIME = 31
QUOTA_SOFT_TIMEOUT = 32
QUOTA_HARD_TIMEOUT = 33
QUOTA_DEFAULT_SOFT_LIMIT = 34
SNAPSHOT_CREATED = 35
SNAPSHOT_CREATE_FAILED = 36
SNAPSHOT_ACTIVATED = 37
SNAPSHOT_ACTIVATE_FAILED = 38
SNAPSHOT_DEACTIVATED = 39
SNAPSHOT_DEACTIVATE_FAILED = 40
SNAPSHOT_SOFT_LIMIT_REACHED = 41
SNAPSHOT_HARD_LIMIT_REACHED = 42
SNAPSHOT_RESTORED = 43
SNAPSHOT_RESTORE_FAILED = 44
SNAPSHOT_DELETED = 45
SNAPSHOT_DELETE_FAILED = 46
SNAPSHOT_CLONED = 47
SNAPSHOT_CLONE_FAILED = 48
SNAPSHOT_CONFIG_UPDATED = 49
SNAPSHOT_CONFIG_UPDATE_FAILED = 50
SNAPSHOT_SCHEDULER_INITIALISED = 51
SNAPSHOT_SCHEDULER_INIT_FAILED = 52
SNAPSHOT_SCHEDULER_ENABLED = 53
SNAPSHOT_SCHEDULER_ENABLE_FAILED = 54
SNAPSHOT_SCHEDULER_DISABLED = 55
SNAPSHOT_SCHEDULER_DISABLE_FAILED = 56
SNAPSHOT_SCHEDULER_SCHEDULE_ADDED = 57
SNAPSHOT_SCHEDULER_SCHEDULE_ADD_FAILED = 58
SNAPSHOT_SCHEDULER_SCHEDULE_EDITED = 59
SNAPSHOT_SCHEDULER_SCHEDULE_EDIT_FAILED = 60
SNAPSHOT_SCHEDULER_SCHEDULE_DELETED = 61
SNAPSHOT_SCHEDULER_SCHEDULE_DELETE_FAILED = 62
SVC_MANAGER_FAILED = 63
SVC_RECONFIGURE_FAILED = 64
SVC_CONNECTED = 65
SVC_DISCONNECTED = 66
PEER_STORE_FAILURE = 67
PEER_RPC_CREATE_FAILED = 68
PEER_REJECT = 69
PEER_CONNECT = 70
PEER_DISCONNECT = 71
PEER_NOT_FOUND = 72
UNKNOWN_PEER = 73
BRICK_START_FAILED = 74
BRICK_STOP_FAILED = 75
BRICK_DISCONNECTED = 76
BRICK_CONNECTED = 77
BRICKPATH_RESOLVE_FAILED = 78
NOTIFY_UNKNOWN_OP = 79
QUORUM_LOST = 80
QUORUM_REGAINED = 81
REBALANCE_START_FAILED = 82
REBALANCE_STATUS_UPDATE_FAILED = 83
IMPORT_QUOTA_CONF_FAILED = 84
IMPORT_VOLUME_FAILED = 85
IMPORT_BRICK_FAILED = 86
COMPARE_FRIEND_VOLUME_FAILED = 87
NFS_GANESHA_EXPORT_FAILED = 88
EC_MIN_BRICKS_NOT_UP = 89
EC_MIN_BRICKS_UP = 90
GEOREP_FAULTY = 91
GEOREP_CHECKPOINT_COMPLETED = 92
GEOREP_ACTIVE = 93
GEOREP_PASSIVE = 94
QUOTA_CROSSED_SOFT_LIMIT = 95
BITROT_BAD_FILE = 96
CLIENT_CONNECT = 97
CLIENT_AUTH_REJECT = 98
CLIENT_DISCONNECT = 99
POSIX_SAME_GFID = 100
POSIX_ALREADY_PART_OF_VOLUME = 101
POSIX_BRICK_NOT_IN_VOLUME = 102
POSIX_BRICK_VERIFICATION_FAILED = 103
POSIX_ACL_NOT_SUPPORTED = 104
POSIX_HEALTH_CHECK_FAILED = 105
AFR_QUORUM_MET = 106
AFR_QUORUM_FAIL = 107
AFR_SUBVOL_UP = 108
AFR_SUBVOLS_DOWN = 109
AFR_SPLIT_BRAIN = 110
TIER_ATTACH = 111
TIER_ATTACH_FORCE = 112
TIER_DETACH_START = 113
TIER_DETACH_STOP = 114
TIER_DETACH_COMMIT = 115
TIER_DETACH_FORCE = 116
TIER_PAUSE = 117
TIER_RESUME = 118
TIER_WATERMARK_HI = 119
TIER_WATERMARK_DROPPED_TO_MID = 120
TIER_WATERMARK_RAISED_TO_MID = 121
TIER_WATERMARK_DROPPED_TO_LOW = 122
VOLUME_ADD_BRICK = 123
VOLUME_ADD_BRICK_FAILED = 124
VOLUME_REMOVE_BRICK_START = 125
VOLUME_REMOVE_BRICK_START_FAILED = 126
VOLUME_REMOVE_BRICK_COMMIT = 127
VOLUME_REMOVE_BRICK_COMMIT_FAILED = 128
VOLUME_REMOVE_BRICK_STOP = 129
VOLUME_REMOVE_BRICK_STOP_FAILED = 130
VOLUME_REMOVE_BRICK_FORCE = 131
VOLUME_REMOVE_BRICK_FORCE_FAILED = 132
VOLUME_REMOVE_BRICK_FAILED = 133
VOLUME_REBALANCE_START = 134
VOLUME_REBALANCE_STOP = 135
VOLUME_REBALANCE_FAILED = 136
VOLUME_REBALANCE_COMPLETE = 137
TIER_START = 138
TIER_START_FORCE = 139
DHT_DISK_USAGE = 140
DHT_INODES_USAGE = 141



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
#  Copyright (c) 2016 Red Hat, Inc. <http://www.redhat.com>
#  This file is part of GlusterFS.
#
#  This file is licensed to you under your choice of the GNU Lesser
#  General Public License, version 3 or any later version (LGPLv3 or
#  later), or the GNU General Public License, version 2 (GPLv2), in all
#  cases as published by the Free Software Foundation.
#

from __future__ import print_function
from argparse import ArgumentParser, RawDescriptionHelpFormatter
import logging
from datetime import datetime

from flask import Flask, request

app = Flask(__name__)
app.logger.disabled = True
log = logging.getLogger('werkzeug')
log.disabled = True


def human_time(ts):
    return datetime.fromtimestamp(float(ts)).strftime("%Y-%m-%d %H:%M:%S")


@app.route("/")
def home():
    return "OK"


@app.route("/listen", methods=["POST"])
def listen():
    data = request.json
    if data is None:
        return "OK"

    message = []
    for k, v in data.get("message", {}).items():
        message.append("{0}={1}".format(k, v))

    print(("{0:20s} {1:20s} {2:36} {3}".format(
        human_time(data.get("ts")),
        data.get("event"),
        data.get("nodeid"),
        " ".join(message))))

    return "OK"


def main():
    parser = ArgumentParser(formatter_class=RawDescriptionHelpFormatter,
                            description=__doc__)
    parser.add_argument("--port", type=int, help="Port", default=9000)
    parser.add_argument("--debug", help="Run Server in debug mode",
                        action="store_true")
    args = parser.parse_args()

    print(("{0:20s} {1:20s} {2:36} {3}".format(
        "TIMESTAMP", "EVENT", "NODE ID", "MESSAGE"
    )))
    print(("{0:20s} {1:20s} {2:36} {3}".format(
        "-"*20, "-"*20, "-"*36, "-"*20
    )))
    if args.debug:
        app.debug = True

    app.run(host="0.0.0.0", port=args.port)


if __name__ == "__main__":
    main()



#!/usr/bin/python3
#
# Copyright (c) 2015 Red Hat, Inc. <http://www.redhat.com>
# This file is part of GlusterFS.
#
# This file is licensed to you under your choice of the GNU Lesser
# General Public License, version 3 or any later version (LGPLv3 or
# later), or the GNU General Public License, version 2 (GPLv2), in all
# cases as published by the Free Software Foundation.

from __future__ import print_function
import subprocess
import os
import os.path
import sys
import time
import logging
import logging.handlers
import fcntl


GCRON_TASKS = "/run/gluster/shared_storage/snaps/glusterfs_snap_cron_tasks"
GCRON_CROND_TASK = "/etc/cron.d/glusterfs_snap_cron_tasks"
GCRON_RELOAD_FLAG = "/var/run/gluster/crond_task_reload_flag"
LOCK_FILE_DIR = "/run/gluster/shared_storage/snaps/lock_files/"
log = logging.getLogger("gcron-logger")
start_time = 0.0


def initLogger(script_name):
    log.setLevel(logging.DEBUG)
    logFormat = "[%(asctime)s %(filename)s:%(lineno)s %(funcName)s] "\
        "%(levelname)s %(message)s"
    formatter = logging.Formatter(logFormat)

    sh = logging.handlers.SysLogHandler()
    sh.setLevel(logging.ERROR)
    sh.setFormatter(formatter)

    process = subprocess.Popen(["gluster", "--print-logdir"],
                               stdout=subprocess.PIPE,
                               universal_newlines=True)
    out, err = process.communicate()
    if process.returncode == 0:
        logfile = os.path.join(out.strip(), script_name[:-3]+".log")

    fh = logging.FileHandler(logfile)
    fh.setLevel(logging.DEBUG)
    fh.setFormatter(formatter)

    log.addHandler(sh)
    log.addHandler(fh)


def takeSnap(volname="", snapname=""):
    success = True
    if volname == "":
        log.debug("No volname given")
        return False
    if snapname == "":
        log.debug("No snapname given")
        return False

    cli = ["gluster",
           "snapshot",
           "create",
           snapname,
           volname]
    log.debug("Running command '%s'", " ".join(cli))

    p = subprocess.Popen(cli, stdout=subprocess.PIPE,
                         stderr=subprocess.PIPE)
    out, err = p.communicate()
    rv = p.returncode

    log.debug("Command '%s' returned '%d'", " ".join(cli), rv)

    if rv:
        log.error("Snapshot of %s failed", volname)
        log.error("Command output:")
        log.error(err)
        success = False
    else:
        log.info("Snapshot of %s successful", volname)

    return success


def doJob(name, lockFile, jobFunc, volname):
    success = True
    try:
        f = os.open(lockFile, os.O_CREAT | os.O_RDWR | os.O_NONBLOCK)
        try:
            fcntl.flock(f, fcntl.LOCK_EX | fcntl.LOCK_NB)
            mtime = os.path.getmtime(lockFile)
            global start_time
            log.debug("%s last modified at %s", lockFile, time.ctime(mtime))
            if mtime < start_time:
                log.debug("Processing job %s", name)
                if jobFunc(volname, name):
                    log.info("Job %s succeeded", name)
                else:
                    log.error("Job %s failed", name)
                    success = False
                os.utime(lockFile, None)
            else:
                log.info("Job %s has been processed already", name)
            fcntl.flock(f, fcntl.LOCK_UN)
        except (OSError, IOError):
            log.info("Job %s is being processed by another agent", name)
        os.close(f)
    except (OSError, IOError) as e:
        log.debug("Failed to open lock file %s : %s", lockFile, e)
        log.error("Failed to process job %s", name)
        success = False

    return success


def main():
    script_name = os.path.basename(__file__)
    initLogger(script_name)
    global start_time
    if sys.argv[1] == "--update":
        if not os.path.exists(GCRON_TASKS):
            # Create a flag in /var/run/gluster which indicates that this
            # node doesn't have access to GCRON_TASKS right now, so that
            # when the mount is available and GCRON_TASKS is available
            # the flag will tell this routine to reload GCRON_CROND_TASK
            try:
                f = os.open(GCRON_RELOAD_FLAG,
                            os.O_CREAT | os.O_NONBLOCK, 0o644)
                os.close(f)
            except OSError as e:
                if errno != EEXIST:
                    log.error("Failed to create %s : %s",
                              GCRON_RELOAD_FLAG, e)
                    output("Failed to create %s. Error: %s"
                           % (GCRON_RELOAD_FLAG, e))
            return

        if not os.path.exists(GCRON_CROND_TASK):
            return

        # As GCRON_TASKS exists now, we should check if GCRON_RELOAD_FLAG
        # also exists. If so we should touch GCRON_CROND_TASK and remove
        # the GCRON_RELOAD_FLAG
        if os.path.exists(GCRON_RELOAD_FLAG):
            try:
                os.remove(GCRON_RELOAD_FLAG);
                process = subprocess.Popen(["touch", "-h", GCRON_CROND_TASK],
                                           stdout=subprocess.PIPE,
                                           stderr=subprocess.PIPE)
                out, err = process.communicate()
                if process.returncode != 0:
                    log.error("Failed to touch %s. Error: %s.",
                              GCRON_CROND_TASK, err)
            except (IOError, OSError) as e:
                log.error("Failed to touch %s. Error: %s.",
                          GCRON_CROND_TASK, e)
            return
        if os.lstat(GCRON_TASKS).st_mtime > \
           os.lstat(GCRON_CROND_TASK).st_mtime:
            try:
                process = subprocess.Popen(["touch", "-h", GCRON_CROND_TASK],
                                           stdout=subprocess.PIPE,
                                           stderr=subprocess.PIPE)
                out, err = process.communicate()
                if process.returncode != 0:
                    log.error("Failed to touch %s. Error: %s.",
                              GCRON_CROND_TASK, err)
            except IOError as e:
                log.error("Failed to touch %s. Error: %s.",
                          GCRON_CROND_TASK, e)
        return

    volname = sys.argv[1]
    jobname = sys.argv[2]
    locking_file = os.path.join(LOCK_FILE_DIR, jobname)
    log.debug("locking_file = %s", locking_file)
    log.debug("volname = %s", volname)
    log.debug("jobname = %s", jobname)

    start_time = int(time.time())

    doJob("Scheduled-" + jobname + "-" + volname, locking_file, takeSnap, volname)


if __name__ == "__main__":
    main()



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
#  Copyright (c) 2016 Red Hat, Inc. <http://www.redhat.com>
#  This file is part of GlusterFS.
#
#  This file is licensed to you under your choice of the GNU Lesser
#  General Public License, version 3 or any later version (LGPLv3 or
#  later), or the GNU General Public License, version 2 (GPLv2), in all
#  cases as published by the Free Software Foundation.
#
"""
Usage:
    gluster-georep-sshkey generate
    or
    gluster-georep-sshkey generate --no-prefix

Generates two SSH keys(one for gsyncd access and other for tar) in all
peer nodes and collects the public keys to the local node where it is
initiated. Adds `command=` prefix to common_secret.pem.pub if `--no-prefix`
argument is not passed.
"""
import os
import glob

from gluster.cliutils import (node_output_ok, execute, execute_in_peers,
                              Cmd, runcli)
from prettytable import PrettyTable


SECRET_PEM = "/var/lib/glusterd/geo-replication/secret.pem"
TAR_SSH_PEM = "/var/lib/glusterd/geo-replication/tar_ssh.pem"
GSYNCD_CMD = 'command="/usr/lib/glusterfs/glusterfs/gsyncd" '
TAR_CMD = 'command="tar ${SSH_ORIGINAL_COMMAND#* }" '
COMMON_SECRET_FILE = "/var/lib/glusterd/geo-replication/common_secret.pem.pub"


class NodeGenCmd(Cmd):
    name = "node-generate"

    def args(self, parser):
        parser.add_argument("no_prefix")

    def run(self, args):
        # Regenerate if secret.pem.pub not exists
        if not os.path.exists(SECRET_PEM + ".pub"):
            # Cleanup old files
            for f in glob.glob(SECRET_PEM + "*"):
                os.remove(f)

            execute(["ssh-keygen", "-N", "", "-f", SECRET_PEM])

        # Regenerate if ssh_tar.pem.pub not exists
        if not os.path.exists(TAR_SSH_PEM + ".pub"):
            # Cleanup old files
            for f in glob.glob(TAR_SSH_PEM + "*"):
                os.remove(f)

            execute(["ssh-keygen", "-N", "", "-f", TAR_SSH_PEM])

        # Add required prefixes if prefix is not "container"
        prefix_secret_pem_pub = ""
        prefix_tar_ssh_pem_pub = ""
        if args.no_prefix != "no-prefix":
            prefix_secret_pem_pub = GSYNCD_CMD
            prefix_tar_ssh_pem_pub = TAR_CMD

        data = {"default_pub": "", "tar_pub": ""}
        with open(SECRET_PEM + ".pub") as f:
            data["default_pub"] = prefix_secret_pem_pub + f.read().strip()

        with open(TAR_SSH_PEM + ".pub") as f:
            data["tar_pub"] = prefix_tar_ssh_pem_pub + f.read().strip()

        node_output_ok(data)


def color_status(value):
    if value in ["UP", "OK"]:
        return "green"
    return "red"


class GenCmd(Cmd):
    name = "generate"

    def args(self, parser):
        parser.add_argument("--no-prefix", help="Do not use prefix in "
                            "generated pub keys", action="store_true")

    def run(self, args):
        prefix = "no-prefix" if args.no_prefix else "."
        out = execute_in_peers("node-generate", [prefix])

        common_secrets = []
        table = PrettyTable(["NODE", "NODE STATUS", "KEYGEN STATUS"])
        table.align["NODE STATUS"] = "r"
        table.align["KEYGEN STATUS"] = "r"
        for p in out:
            if p.ok:
                common_secrets.append(p.output["default_pub"])
                common_secrets.append(p.output["tar_pub"])

            table.add_row([p.hostname,
                           "UP" if p.node_up else "DOWN",
                           "OK" if p.ok else "NOT OK: {0}".format(
                               p.error)])

        with open(COMMON_SECRET_FILE, "w") as f:
            f.write("\n".join(common_secrets) + "\n")

        print (table)


if __name__ == "__main__":
    runcli()



#
# Copyright (c) 2011-2014 Red Hat, Inc. <http://www.redhat.com>
# This file is part of GlusterFS.

# This file is licensed to you under your choice of the GNU Lesser
# General Public License, version 3 or any later version (LGPLv3 or
# later), or the GNU General Public License, version 2 (GPLv2), in all
# cases as published by the Free Software Foundation.
#

import os
import sys
import pwd
import time
import fcntl
import shutil
import logging
import errno
import threading
import subprocess
import socket
from subprocess import PIPE
from threading import Lock, Thread as baseThread
from errno import (EACCES, EAGAIN, EPIPE, ENOTCONN, ENOMEM, ECONNABORTED,
                   EINTR, ENOENT, ESTALE, EBUSY, ENODATA, errorcode, EIO)
from signal import signal, SIGTERM
import select as oselect
from os import waitpid as owaitpid
import xml.etree.ElementTree as XET
from select import error as SelectError
try:
    from cPickle import PickleError
except ImportError:
    from pickle import PickleError

from conf import GLUSTERFS_LIBEXECDIR, UUID_FILE
sys.path.insert(1, GLUSTERFS_LIBEXECDIR)
EVENTS_ENABLED = True
try:
    from gfevents.eventtypes import GEOREP_FAULTY as EVENT_GEOREP_FAULTY
    from gfevents.eventtypes import GEOREP_ACTIVE as EVENT_GEOREP_ACTIVE
    from gfevents.eventtypes import GEOREP_PASSIVE as EVENT_GEOREP_PASSIVE
    from gfevents.eventtypes import GEOREP_CHECKPOINT_COMPLETED \
        as EVENT_GEOREP_CHECKPOINT_COMPLETED
except ImportError:
    # Events APIs not installed, dummy eventtypes with None
    EVENTS_ENABLED = False
    EVENT_GEOREP_FAULTY = None
    EVENT_GEOREP_ACTIVE = None
    EVENT_GEOREP_PASSIVE = None
    EVENT_GEOREP_CHECKPOINT_COMPLETED = None

import gsyncdconfig as gconf
from rconf import rconf

from hashlib import sha256 as sha256

ENOTSUP = getattr(errno, 'ENOTSUP', 'EOPNOTSUPP')

# auxiliary gfid based access prefix
_CL_AUX_GFID_PFX = ".gfid/"
ROOT_GFID = "00000000-0000-0000-0000-000000000001"
GF_OP_RETRIES = 10

GX_GFID_CANONICAL_LEN = 37  # canonical gfid len + '\0'

NodeID = None
rsync_version = None
unshare_mnt_propagation = None
slv_bricks = None
SPACE_ESCAPE_CHAR = "%20"
NEWLINE_ESCAPE_CHAR = "%0A"
PERCENTAGE_ESCAPE_CHAR = "%25"

final_lock = Lock()

def sup(x, *a, **kw):
    """a rubyesque "super" for python ;)

    invoke caller method in parent class with given args.
    """
    return getattr(super(type(x), x),
                   sys._getframe(1).f_code.co_name)(*a, **kw)


def escape(s):
    """the chosen flavor of string escaping, used all over
       to turn whatever data to creatable representation"""
    return s.replace("/", "-").strip("-")


def escape_space_newline(s):
    return s.replace("%", PERCENTAGE_ESCAPE_CHAR)\
            .replace(" ", SPACE_ESCAPE_CHAR)\
            .replace("\n", NEWLINE_ESCAPE_CHAR)


def unescape_space_newline(s):
    return s.replace(SPACE_ESCAPE_CHAR, " ")\
            .replace(NEWLINE_ESCAPE_CHAR, "\n")\
            .replace(PERCENTAGE_ESCAPE_CHAR, "%")

# gf_mount_ready() returns 1 if all subvols are up, else 0
def gf_mount_ready():
    ret = errno_wrap(Xattr.lgetxattr,
                     ['.', 'dht.subvol.status', 16],
                     [ENOENT, ENOTSUP, ENODATA], [ENOMEM])

    if isinstance(ret, int):
       logging.error("failed to get the xattr value")
       return 1
    ret = ret.rstrip('\x00')
    if ret == "1":
       return 1
    return 0

def norm(s):
    if s:
        return s.replace('-', '_')


def update_file(path, updater, merger=lambda f: True):
    """update a file in a transaction-like manner"""

    fr = fw = None
    try:
        fd = os.open(path, os.O_CREAT | os.O_RDWR)
        try:
            fr = os.fdopen(fd, 'r+b')
        except:
            os.close(fd)
            raise
        fcntl.lockf(fr, fcntl.LOCK_EX)
        if not merger(fr):
            return

        tmpp = path + '.tmp.' + str(os.getpid())
        fd = os.open(tmpp, os.O_CREAT | os.O_EXCL | os.O_WRONLY)
        try:
            fw = os.fdopen(fd, 'wb', 0)
        except:
            os.close(fd)
            raise
        updater(fw)
        os.fsync(fd)
        os.rename(tmpp, path)
    finally:
        for fx in (fr, fw):
            if fx:
                fx.close()


def create_manifest(fname, content):
    """
    Create manifest file for SSH Control Path
    """
    fd = None
    try:
        fd = os.open(fname, os.O_CREAT | os.O_RDWR)
        try:
            os.write(fd, content)
        except:
            os.close(fd)
            raise
    finally:
        if fd is not None:
            os.close(fd)


def setup_ssh_ctl(ctld, remote_addr, resource_url):
    """
    Setup GConf ssh control path parameters
    """
    rconf.ssh_ctl_dir = ctld
    content = "SECONDARY_HOST=%s\nSECONDARY_RESOURCE_URL=%s" % (remote_addr,
                                                        resource_url)
    encoded_content = content.encode()
    content_sha256 = sha256hex(encoded_content)
    """
    The length of ctl_path for ssh connection should not be > 108.
    ssh fails with ctl_path too long if it is so. But when rsync
    is piped to ssh, it is not taking > 90. Hence using first 32
    bytes of hash. Hash collision doesn't matter as only one sock
    file is created per directory.
    """
    content_sha256 = content_sha256[:32]
    fname = os.path.join(rconf.ssh_ctl_dir,
                         "%s.mft" % content_sha256)

    create_manifest(fname, encoded_content)
    ssh_ctl_path = os.path.join(rconf.ssh_ctl_dir,
                                "%s.sock" % content_sha256)
    rconf.ssh_ctl_args = ["-oControlMaster=auto", "-S", ssh_ctl_path]


def grabfile(fname, content=None):
    """open @fname + contest for its fcntl lock

    @content: if given, set the file content to it
    """
    # damn those messy open() mode codes
    fd = os.open(fname, os.O_CREAT | os.O_RDWR)
    f = os.fdopen(fd, 'r+')
    try:
        fcntl.lockf(f, fcntl.LOCK_EX | fcntl.LOCK_NB)
    except:
        ex = sys.exc_info()[1]
        f.close()
        if isinstance(ex, IOError) and ex.errno in (EACCES, EAGAIN):
            # cannot grab, it's taken
            return
        raise
    if content:
        try:
            f.truncate()
            f.write(content)
            f.flush()
        except:
            f.close()
            raise
    rconf.permanent_handles.append(f)
    return f


def grabpidfile(fname=None, setpid=True):
    """.grabfile customization for pid files"""
    if not fname:
        fname = gconf.get("pid-file")
    content = None
    if setpid:
        content = str(os.getpid()) + '\n'
    return grabfile(fname, content=content)


def finalize(*args, **kwargs):
    """all those messy final steps we go trough upon termination

    Do away with pidfile, ssh control dir and logging.
    """

    final_lock.acquire()
    if gconf.get('pid_file'):
        rm_pidf = rconf.pid_file_owned
        if rconf.cpid:
            # exit path from parent branch of daemonization
            rm_pidf = False
            while True:
                f = grabpidfile(setpid=False)
                if not f:
                    # child has already taken over pidfile
                    break
                if os.waitpid(rconf.cpid, os.WNOHANG)[0] == rconf.cpid:
                    # child has terminated
                    rm_pidf = True
                    break
                time.sleep(0.1)
        if rm_pidf:
            try:
                os.unlink(rconf.pid_file)
            except:
                ex = sys.exc_info()[1]
                if ex.errno == ENOENT:
                    pass
                else:
                    raise
    if rconf.ssh_ctl_dir and not rconf.cpid:
        def handle_rm_error(func, path, exc_info):
            if exc_info[1].errno == ENOENT:
                return
            raise exc_info[1]

        shutil.rmtree(rconf.ssh_ctl_dir, onerror=handle_rm_error)

    """ Unmount if not done """
    if rconf.mount_point:
        if rconf.mountbroker:
            umount_cmd = rconf.mbr_umount_cmd + [rconf.mount_point, 'lazy']
        else:
            umount_cmd = ['umount', '-l', rconf.mount_point]
        p0 = subprocess.Popen(umount_cmd, stderr=subprocess.PIPE,
                              universal_newlines=True)
        _, errdata = p0.communicate()
        if p0.returncode == 0:
            try:
                os.rmdir(rconf.mount_point)
            except OSError:
                pass
        else:
            pass

    if rconf.log_exit:
        logging.info("exiting.")
    sys.stdout.flush()
    sys.stderr.flush()
    os._exit(kwargs.get('exval', 0))


def log_raise_exception(excont):
    """top-level exception handler

    Try to some fancy things to cover up we face with an error.
    Translate some weird sounding but well understood exceptions
    into human-friendly lingo
    """

    is_filelog = False
    for h in logging.getLogger().handlers:
        fno = getattr(getattr(h, 'stream', None), 'fileno', None)
        if fno and not os.isatty(fno()):
            is_filelog = True

    exc = sys.exc_info()[1]
    if isinstance(exc, SystemExit):
        excont.exval = exc.code or 0
        raise
    else:
        logtag = None
        if isinstance(exc, GsyncdError):
            if is_filelog:
                logging.error(exc.args[0])
            sys.stderr.write('failure: ' + exc.args[0] + '\n')
        elif isinstance(exc, PickleError) or isinstance(exc, EOFError) or \
            ((isinstance(exc, OSError) or isinstance(exc, IOError)) and
             exc.errno == EPIPE):
            logging.error('connection to peer is broken')
            if hasattr(rconf, 'transport'):
                rconf.transport.wait()
                if rconf.transport.returncode == 127:
                    logging.error("getting \"No such file or directory\""
                                  "errors is most likely due to "
                                  "MISCONFIGURATION, please remove all "
                                  "the public keys added by geo-replication "
                                  "from authorized_keys file in secondary nodes "
                                  "and run Geo-replication create "
                                  "command again.")
                    logging.error("If `gsec_create container` was used, then "
                                  "run `gluster volume geo-replication "
                                  "<PRIMARYVOL> [<SECONDARYUSER>@]<SECONDARYHOST>::"
                                  "<SECONDARYVOL> config remote-gsyncd "
                                  "<GSYNCD_PATH> (Example GSYNCD_PATH: "
                                  "`/usr/libexec/glusterfs/gsyncd`)")
                rconf.transport.terminate_geterr()
        elif isinstance(exc, OSError) and exc.errno in (ENOTCONN,
                                                        ECONNABORTED):
            logging.error(lf('Gluster Mount process exited',
                             error=errorcode[exc.errno]))
        elif isinstance(exc, OSError) and exc.errno == EIO:
            logging.error("Getting \"Input/Output error\" "
                          "is most likely due to "
                          "a. Brick is down or "
                          "b. Split brain issue.")
            logging.error("This is expected as per design to "
                          "keep the consistency of the file system. "
                          "Once the above issue is resolved "
                          "geo-replication would automatically "
                          "proceed further.")
            logtag = "FAIL"
        else:
            logtag = "FAIL"
        if not logtag and logging.getLogger().isEnabledFor(logging.DEBUG):
            logtag = "FULL EXCEPTION TRACE"
        if logtag:
            logging.exception(logtag + ": ")
            sys.stderr.write("failed with %s: %s.\n" % (type(exc).__name__, exc))
        excont.exval = 1
        sys.exit(excont.exval)


class FreeObject(object):

    """wildcard class for which any attribute can be set"""

    def __init__(self, **kw):
        for k, v in kw.items():
            setattr(self, k, v)


class Thread(baseThread):

    """thread class flavor for gsyncd

    - always a daemon thread
    - force exit for whole program if thread
      function coughs up an exception
    """

    def __init__(self, *args, **kwargs):
        tf = kwargs.get('target')
        if tf:
            def twrap(*aargs):
                excont = FreeObject(exval=0)
                try:
                    tf(*aargs)
                except:
                    try:
                        log_raise_exception(excont)
                    finally:
                        finalize(exval=excont.exval)
            kwargs['target'] = twrap
        baseThread.__init__(self, *args, **kwargs)
        self.setDaemon(True)


class GsyncdError(Exception):
    pass


class _MetaXattr(object):

    """singleton class, a lazy wrapper around the
    libcxattr module

    libcxattr (a heavy import due to ctypes) is
    loaded only when when the single
    instance is tried to be used.

    This reduces runtime for those invocations
    which do not need filesystem manipulation
    (eg. for config, url parsing)
    """

    def __getattr__(self, meth):
        from libcxattr import Xattr as LXattr
        xmeth = [m for m in dir(LXattr) if m[0] != '_']
        if meth not in xmeth:
            return
        for m in xmeth:
            setattr(self, m, getattr(LXattr, m))
        return getattr(self, meth)


Xattr = _MetaXattr()


def getusername(uid=None):
    if uid is None:
        uid = os.geteuid()
    return pwd.getpwuid(uid).pw_name


def privileged():
    return os.geteuid() == 0


def boolify(s):
    """
    Generic string to boolean converter

    return
    - Quick return if string 's' is of type bool
    - True if it's in true_list
    - False if it's in false_list
    - Warn if it's not present in either and return False
    """
    true_list = ['true', 'yes', '1', 'on']
    false_list = ['false', 'no', '0', 'off']

    if isinstance(s, bool):
        return s

    rv = False
    lstr = s.lower()
    if lstr in true_list:
        rv = True
    elif lstr not in false_list:
        logging.warn(lf("Unknown string in \"string to boolean\" conversion, "
                        "defaulting to False",
                        str=s))

    return rv


def eintr_wrap(func, exc, *args):
    """
    wrapper around syscalls resilient to interrupt caused
    by signals
    """
    while True:
        try:
            return func(*args)
        except exc:
            ex = sys.exc_info()[1]
            if not ex.args[0] == EINTR:
                raise


def select(*args):
    return eintr_wrap(oselect.select, oselect.error, *args)


def waitpid(*args):
    return eintr_wrap(owaitpid, OSError, *args)


def term_handler_default_hook(signum, frame):
    finalize(signum, frame, exval=1)


def set_term_handler(hook=term_handler_default_hook):
    signal(SIGTERM, hook)


def get_node_uuid():
    global NodeID
    if NodeID is not None:
        return NodeID

    NodeID = ""
    with open(UUID_FILE) as f:
        for line in f:
            if line.startswith("UUID="):
                NodeID = line.strip().split("=")[-1]
                break

    if NodeID == "":
        raise GsyncdError("Failed to get Host UUID from %s" % UUID_FILE)
    return NodeID


def is_host_local(host_id):
    return host_id == get_node_uuid()


def funcode(f):
    fc = getattr(f, 'func_code', None)
    if not fc:
        # python 3
        fc = f.__code__
    return fc


def memoize(f):
    fc = funcode(f)
    fn = fc.co_name

    def ff(self, *a, **kw):
        rv = getattr(self, '_' + fn, None)
        if rv is None:
            rv = f(self, *a, **kw)
            setattr(self, '_' + fn, rv)
        return rv
    return ff


def umask():
    return os.umask(0)


def entry2pb(e):
    return e.rsplit('/', 1)


def gauxpfx():
    return _CL_AUX_GFID_PFX


def sha256hex(s):
    return sha256(s).hexdigest()


def selfkill(sig=SIGTERM):
    os.kill(os.getpid(), sig)


def errno_wrap(call, arg=[], errnos=[], retry_errnos=[]):
    """ wrapper around calls resilient to errnos.
    """
    nr_tries = 0
    while True:
        try:
            return call(*arg)
        except OSError:
            ex = sys.exc_info()[1]
            if ex.errno in errnos:
                return ex.errno
            if ex.errno not in retry_errnos:
                raise
            nr_tries += 1
            if nr_tries == GF_OP_RETRIES:
                # probably a screwed state, cannot do much...
                logging.warn(lf('reached maximum retries',
                                args=repr(arg),
                                error=ex))
                raise
            time.sleep(0.250)  # retry the call


def lstat(e):
    return errno_wrap(os.lstat, [e], [ENOENT], [ESTALE, EBUSY])

def get_gfid_from_mnt(gfidpath):
    return errno_wrap(Xattr.lgetxattr,
                      [gfidpath, 'glusterfs.gfid.string',
                       GX_GFID_CANONICAL_LEN], [ENOENT], [ESTALE])


def matching_disk_gfid(gfid, entry):
    disk_gfid = get_gfid_from_mnt(entry)
    if isinstance(disk_gfid, int):
        return False

    if not gfid == disk_gfid:
        return False

    return True


class NoStimeAvailable(Exception):
    pass


class PartialHistoryAvailable(Exception):
    pass


class ChangelogHistoryNotAvailable(Exception):
    pass


class ChangelogException(OSError):
    pass


def gf_event(event_type, **kwargs):
    if EVENTS_ENABLED:
        from gfevents.gf_event import gf_event as gfevent
        gfevent(event_type, **kwargs)


class GlusterLogLevel(object):
        NONE = 0
        EMERG = 1
        ALERT = 2
        CRITICAL = 3
        ERROR = 4
        WARNING = 5
        NOTICE = 6
        INFO = 7
        DEBUG = 8
        TRACE = 9


def get_changelog_log_level(lvl):
    return getattr(GlusterLogLevel, lvl, GlusterLogLevel.INFO)


def get_primary_and_secondary_data_from_args(args):
    primary_name = None
    secondary_data = None
    for arg in args:
        if arg.startswith(":"):
            primary_name = arg.replace(":", "")
        if "::" in arg:
            secondary_data = arg.replace("ssh://", "")

    return (primary_name, secondary_data)

def unshare_propagation_supported():
    global unshare_mnt_propagation
    if unshare_mnt_propagation is not None:
        return unshare_mnt_propagation

    unshare_mnt_propagation = False
    p = subprocess.Popen(["unshare", "--help"],
                         stderr=subprocess.PIPE,
                         stdout=subprocess.PIPE,
                         universal_newlines=True)
    out, err = p.communicate()
    if p.returncode == 0:
        if "propagation" in out:
            unshare_mnt_propagation = True

    return unshare_mnt_propagation


def get_rsync_version(rsync_cmd):
    global rsync_version
    if rsync_version is not None:
        return rsync_version

    rsync_version = "0"
    p = subprocess.Popen([rsync_cmd, "--version"],
                         stderr=subprocess.PIPE,
                         stdout=subprocess.PIPE,
                         universal_newlines=True)
    out, err = p.communicate()
    if p.returncode == 0:
        rsync_version = out.split(" ", 4)[3]

    return rsync_version


def get_slv_dir_path(slv_host, slv_volume, gfid):
    global slv_bricks

    dir_path = ENOENT
    pfx = gauxpfx()

    if not slv_bricks:
        slv_info = Volinfo(slv_volume, slv_host, primary=False)
        slv_bricks = slv_info.bricks
    # Result of readlink would be of format as below.
    # readlink = "../../pgfid[0:2]/pgfid[2:4]/pgfid/basename"
    for brick in slv_bricks:
        dir_path = errno_wrap(os.path.join,
                              [brick['dir'],
                               ".glusterfs", gfid[0:2],
                               gfid[2:4],
                               gfid], [ENOENT], [ESTALE])
        if dir_path != ENOENT:
            try:
                realpath = errno_wrap(os.readlink, [dir_path],
                                      [ENOENT], [ESTALE])
                if not isinstance(realpath, int):
                    realpath_parts = realpath.split('/')
                    pargfid = realpath_parts[-2]
                    basename = realpath_parts[-1]
                    dir_entry = os.path.join(pfx, pargfid, basename)
                    return dir_entry
            except OSError:
                # .gfid/GFID
                gfidpath = unescape_space_newline(os.path.join(pfx, gfid))
                realpath = errno_wrap(Xattr.lgetxattr_buf,
                      [gfidpath, 'glusterfs.gfid2path'], [ENOENT], [ESTALE])
                if not isinstance(realpath, int):
                    basename = os.path.basename(realpath).rstrip('\x00')
                    dirpath = os.path.dirname(realpath)
                    if dirpath == "/":
                        pargfid = ROOT_GFID
                    else:
                        dirpath = dirpath.strip("/")
                        pargfid = get_gfid_from_mnt(dirpath)
                        if isinstance(pargfid, int):
                            return None
                    dir_entry = os.path.join(pfx, pargfid, basename)
                    return dir_entry

    return None


def lf(event, **kwargs):
    """
    Log Format helper function, log messages can be
    easily modified to structured log format.
    lf("Config Change", sync_jobs=4, brick=/bricks/b1) will be
    converted as "Config Change [{brick=/bricks/b1}, {sync_jobs=4}]"
    """
    msgparts = []
    for k, v in kwargs.items():
        msgparts.append("{%s=%s}" % (k, v))
    return "%s [%s]" % (event, ", ".join(msgparts))


class Popen(subprocess.Popen):

    """customized subclass of subprocess.Popen with a ring
    buffer for children error output"""

    @classmethod
    def init_errhandler(cls):
        """start the thread which handles children's error output"""
        cls.errstore = {}

        def tailer():
            while True:
                errstore = cls.errstore.copy()
                try:
                    poe, _, _ = select(
                        [po.stderr for po in errstore], [], [], 1)
                except (ValueError, SelectError):
                    # stderr is already closed wait for some time before
                    # checking next error
                    time.sleep(0.5)
                    continue
                for po in errstore:
                    if po.stderr not in poe:
                        continue
                    po.lock.acquire()
                    try:
                        if po.on_death_row:
                            continue
                        la = errstore[po]
                        try:
                            fd = po.stderr.fileno()
                        except ValueError:  # file is already closed
                            time.sleep(0.5)
                            continue

                        try:
                            l = os.read(fd, 1024)
                        except OSError:
                            time.sleep(0.5)
                            continue

                        if not l:
                            continue
                        tots = len(l)
                        for lx in la:
                            tots += len(lx)
                        while tots > 1 << 20 and la:
                            tots -= len(la.pop(0))
                        la.append(l)
                    finally:
                        po.lock.release()
        t = Thread(target=tailer)
        t.start()
        cls.errhandler = t

    @classmethod
    def fork(cls):
        """fork wrapper that restarts errhandler thread in child"""
        pid = os.fork()
        if not pid:
            cls.init_errhandler()
        return pid

    def __init__(self, args, *a, **kw):
        """customizations for subprocess.Popen instantiation

        - 'close_fds' is taken to be the default
        - if child's stderr is chosen to be managed,
          register it with the error handler thread
        """
        self.args = args
        if 'close_fds' not in kw:
            kw['close_fds'] = True
        self.lock = threading.Lock()
        self.on_death_row = False
        self.elines = []
        try:
            sup(self, args, *a, **kw)
        except:
            ex = sys.exc_info()[1]
            if not isinstance(ex, OSError):
                raise
            raise GsyncdError("""execution of "%s" failed with %s (%s)""" %
                              (args[0], errno.errorcode[ex.errno],
                               os.strerror(ex.errno)))
        if kw.get('stderr') == subprocess.PIPE:
            assert(getattr(self, 'errhandler', None))
            self.errstore[self] = []

    def errlog(self):
        """make a log about child's failure event"""
        logging.error(lf("command returned error",
                         cmd=" ".join(self.args),
                         error=self.returncode))
        lp = ''

        def logerr(l):
            logging.error(self.args[0] + "> " + l)
        for l in self.elines:
            ls = l.split(b'\n')
            ls[0] = lp + ls[0]
            lp = ls.pop()
            for ll in ls:
                logerr(ll)
        if lp:
            logerr(lp)

    def errfail(self):
        """fail nicely if child did not terminate with success"""
        self.errlog()
        finalize(exval=1)

    def terminate_geterr(self, fail_on_err=True):
        """kill child, finalize stderr harvesting (unregister
        from errhandler, set up .elines), fail on error if
        asked for
        """
        self.lock.acquire()
        try:
            self.on_death_row = True
        finally:
            self.lock.release()
        elines = self.errstore.pop(self)
        if self.poll() is None:
            self.terminate()
            if self.poll() is None:
                time.sleep(0.1)
                self.kill()
                self.wait()
        while True:
            if not select([self.stderr], [], [], 0.1)[0]:
                break
            b = os.read(self.stderr.fileno(), 1024)
            if b:
                elines.append(b.decode())
            else:
                break
        self.stderr.close()
        self.elines = elines
        if fail_on_err and self.returncode != 0:
            self.errfail()


def host_brick_split(value):
    """
    IPv6 compatible way to split and get the host
    and brick information. Example inputs:
    node1.example.com:/exports/bricks/brick1/brick
    fe80::af0f:df82:844f:ef66%utun0:/exports/bricks/brick1/brick
    """
    parts = value.split(":")
    brick = parts[-1]
    hostparts = parts[0:-1]
    return (":".join(hostparts), brick)


class Volinfo(object):

    def __init__(self, vol, host='localhost', prelude=[], primary=True):
        if primary:
            gluster_cmd_dir = gconf.get("gluster-command-dir")
        else:
            gluster_cmd_dir = gconf.get("secondary-gluster-command-dir")

        gluster_cmd = os.path.join(gluster_cmd_dir, 'gluster')
        po = Popen(prelude + [gluster_cmd, '--xml', '--remote-host=' + host,
                              'volume', 'info', vol],
                   stdout=PIPE, stderr=PIPE, universal_newlines=True)
        vix = po.stdout.read()
        po.wait()
        po.terminate_geterr()
        vi = XET.fromstring(vix)
        if vi.find('opRet').text != '0':
            if prelude:
                via = '(via %s) ' % ' '.join(prelude)
            else:
                via = ' '
            raise GsyncdError('getting volume info of %s%s '
                              'failed with errorcode %s' %
                              (vol, via, vi.find('opErrno').text))
        self.tree = vi
        self.volume = vol
        self.host = host

    def get(self, elem):
        return self.tree.findall('.//' + elem)

    def is_tier(self):
        return (self.get('typeStr')[0].text == 'Tier')

    def is_hot(self, brickpath):
        logging.debug('brickpath: ' + repr(brickpath))
        return brickpath in self.hot_bricks

    @property
    @memoize
    def bricks(self):
        def bparse(b):
            host, dirp = host_brick_split(b.find("name").text)
            return {'host': host, 'dir': dirp, 'uuid': b.find("hostUuid").text}
        return [bparse(b) for b in self.get('brick')]

    @property
    @memoize
    def uuid(self):
        ids = self.get('id')
        if len(ids) != 1:
            raise GsyncdError("volume info of %s obtained from %s: "
                              "ambiguous uuid" % (self.volume, self.host))
        return ids[0].text

    def replica_count(self, tier, hot):
        if (tier and hot):
            return int(self.get('hotBricks/hotreplicaCount')[0].text)
        elif (tier and not hot):
            return int(self.get('coldBricks/coldreplicaCount')[0].text)
        else:
            return int(self.get('replicaCount')[0].text)

    def disperse_count(self, tier, hot):
        if (tier and hot):
            # Tiering doesn't support disperse volume as hot brick,
            # hence no xml output, so returning 0. In case, if it's
            # supported later, we should change here.
            return 0
        elif (tier and not hot):
            return int(self.get('coldBricks/colddisperseCount')[0].text)
        else:
            return int(self.get('disperseCount')[0].text)

    def distribution_count(self, tier, hot):
        if (tier and hot):
            return int(self.get('hotBricks/hotdistCount')[0].text)
        elif (tier and not hot):
            return int(self.get('coldBricks/colddistCount')[0].text)
        else:
            return int(self.get('distCount')[0].text)

    @property
    @memoize
    def hot_bricks(self):
        return [b.text for b in self.get('hotBricks/brick')]

    def get_hot_bricks_count(self, tier):
        if (tier):
            return int(self.get('hotBricks/hotbrickCount')[0].text)
        else:
            return 0


class VolinfoFromGconf(object):
    # Glusterd will generate following config items before Geo-rep start
    # So that Geo-rep need not run gluster commands from inside
    # Volinfo object API/interface kept as is so that caller need not
    # change anything except calling this instead of Volinfo()
    #
    # primary-bricks=
    # primary-bricks=NODEID:HOSTNAME:PATH,..
    # secondary-bricks=NODEID:HOSTNAME,..
    # primary-volume-id=
    # secondary-volume-id=
    # primary-replica-count=
    # primary-disperse_count=
    def __init__(self, vol, host='localhost', primary=True):
        self.volume = vol
        self.host = host
        self.primary = primary

    def is_tier(self):
        return False

    def is_hot(self, brickpath):
        return False

    def is_uuid(self, value):
        try:
            uuid.UUID(value)
            return True
        except ValueError:
            return False

    def possible_path(self, value):
        return "/" in value

    @property
    @memoize
    def bricks(self):
        pfx = "primary-" if self.primary else "secondary-"
        bricks_data = gconf.get(pfx + "bricks")
        if bricks_data is None:
            return []

        bricks_data = bricks_data.split(",")
        bricks_data = [b.strip() for b in bricks_data]
        out = []
        for b in bricks_data:
            parts = b.split(":")
            b_uuid = None
            if self.is_uuid(parts[0]):
                b_uuid = parts[0]
                # Set all parts except first
                parts = parts[1:]

            if self.possible_path(parts[-1]):
                bpath = parts[-1]
                # Set all parts except last
                parts = parts[0:-1]

            out.append({
                "host": ":".join(parts),   # if remaining parts are IPv6 name
                "dir": bpath,
                "uuid": b_uuid
            })

        return out

    @property
    @memoize
    def uuid(self):
        if self.primary:
            return gconf.get("primary-volume-id")
        else:
            return gconf.get("secondary-volume-id")

    def replica_count(self, tier, hot):
        return gconf.get("primary-replica-count")

    def disperse_count(self, tier, hot):
        return gconf.get("primary-disperse-count")

    def distribution_count(self, tier, hot):
        return gconf.get("primary-distribution-count")

    @property
    @memoize
    def hot_bricks(self):
        return []

    def get_hot_bricks_count(self, tier):
        return 0


def can_ssh(host, port=22):
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    try:
        s.connect((host, port))
        flag = True
    except socket.error:
        flag = False

    s.close()
    return flag


def get_up_nodes(hosts, port):
    # List of hosts with Hostname/IP and UUID
    up_nodes = []
    for h in hosts:
        if can_ssh(h[0], port):
            up_nodes.append(h)

    return up_nodes


def ssh_cipher_present(sshopts):
    """
    Returns True if user has defined a custom cipher to
    use for the SSH connection under rysnc-ssh-options.
    """
    if not sshopts:
        return False

    for opt in sshopts:
        if opt[:2] == "-c":
            return True

    return False



#
# Copyright (c) 2011-2014 Red Hat, Inc. <http://www.redhat.com>
# This file is part of GlusterFS.

# This file is licensed to you under your choice of the GNU Lesser
# General Public License, version 3 or any later version (LGPLv3 or
# later), or the GNU General Public License, version 2 (GPLv2), in all
# cases as published by the Free Software Foundation.
#

import os
from ctypes import CDLL, RTLD_GLOBAL, get_errno, byref, c_ulong
from ctypes.util import find_library
from syncdutils import ChangelogException, ChangelogHistoryNotAvailable
from py2py3 import (gr_cl_history_changelog, gr_cl_done,
                    gr_create_string_buffer, gr_cl_register,
                    gr_cl_history_done, bytearray_to_str)


libgfc = CDLL(
    find_library("gfchangelog"),
    mode=RTLD_GLOBAL,
    use_errno=True
)


def _raise_changelog_err():
    errn = get_errno()
    raise ChangelogException(errn, os.strerror(errn))


def _init():
    if libgfc.gf_changelog_init(None) == -1:
        _raise_changelog_err()


def register(brick, path, log_file, log_level, retries=0):
    _init()

    ret = gr_cl_register(libgfc, brick, path, log_file, log_level, retries)

    if ret == -1:
        _raise_changelog_err()


def scan():
    ret = libgfc.gf_changelog_scan()
    if ret == -1:
        _raise_changelog_err()


def startfresh():
    ret = libgfc.gf_changelog_start_fresh()
    if ret == -1:
        _raise_changelog_err()


def getchanges():
    def clsort(cfile):
        return cfile.split('.')[-1]

    changes = []
    buf = gr_create_string_buffer(4096)
    call = libgfc.gf_changelog_next_change

    while True:
        ret = call(buf, 4096)
        if ret in (0, -1):
            break

        # py2 and py3 compatibility
        result = bytearray_to_str(buf.raw[:ret - 1])
        changes.append(result)

    if ret == -1:
        _raise_changelog_err()

    # cleanup tracker
    startfresh()

    return sorted(changes, key=clsort)


def done(clfile):
    ret = gr_cl_done(libgfc, clfile)
    if ret == -1:
        _raise_changelog_err()


def history_scan():
    ret = libgfc.gf_history_changelog_scan()
    if ret == -1:
        _raise_changelog_err()

    return ret


def history_changelog(changelog_path, start, end, num_parallel):
    actual_end = c_ulong()
    ret = gr_cl_history_changelog(libgfc, changelog_path, start, end,
                                  num_parallel, byref(actual_end))
    if ret == -1:
        _raise_changelog_err()

    if ret == -2:
        raise ChangelogHistoryNotAvailable()

    return (ret, actual_end.value)


def history_startfresh():
    ret = libgfc.gf_history_changelog_start_fresh()
    if ret == -1:
        _raise_changelog_err()


def history_getchanges():
    def clsort(cfile):
        return cfile.split('.')[-1]

    changes = []
    buf = gr_create_string_buffer(4096)
    call = libgfc.gf_history_changelog_next_change

    while True:
        ret = call(buf, 4096)
        if ret in (0, -1):
            break

        # py2 and py3 compatibility
        result = bytearray_to_str(buf.raw[:ret - 1])
        changes.append(result)

    if ret == -1:
        _raise_changelog_err()

    return sorted(changes, key=clsort)


def history_done(clfile):
    ret = gr_cl_history_done(libgfc, clfile)
    if ret == -1:
        _raise_changelog_err()



#!/usr/bin/python3
# -*- coding: utf-8 -*-

# Copyright (c) 2015 Red Hat, Inc. <http://www.redhat.com/>
# This file is part of GlusterFS.
#
# This file is licensed to you under your choice of the GNU Lesser
# General Public License, version 3 or any later version (LGPLv3 or
# later), or the GNU General Public License, version 2 (GPLv2), in all
# cases as published by the Free Software Foundation.

import sys
from errno import ENOENT, ENOTEMPTY
import time
from multiprocessing import Process
import os
import xml.etree.cElementTree as etree
from argparse import ArgumentParser, RawDescriptionHelpFormatter, Action
from gfind_py2py3 import gfind_write_row, gfind_write
import logging
import shutil
import tempfile
import signal
from datetime import datetime
import codecs
import re

from utils import execute, is_host_local, mkdirp, fail
from utils import setup_logger, human_time, handle_rm_error
from utils import get_changelog_rollover_time, cache_output, create_file
import conf
from changelogdata import OutputMerger

PROG_DESCRIPTION = """
GlusterFS Incremental API
"""
ParseError = etree.ParseError if hasattr(etree, 'ParseError') else SyntaxError

logger = logging.getLogger()
vol_statusStr = ""
gtmpfilename = None
g_pid_nodefile_map = {}


class StoreAbsPath(Action):
    def __init__(self, option_strings, dest, nargs=None, **kwargs):
        super(StoreAbsPath, self).__init__(option_strings, dest, **kwargs)

    def __call__(self, parser, namespace, values, option_string=None):
        setattr(namespace, self.dest, os.path.abspath(values))


def get_pem_key_path(session, volume):
    return os.path.join(conf.get_opt("session_dir"),
                        session,
                        volume,
                        "%s_%s_secret.pem" % (session, volume))


def node_cmd(host, host_uuid, task, cmd, args, opts):
    """
    Runs command via ssh if host is not local
    """
    try:
        localdir = is_host_local(host_uuid)

        # this is so to avoid deleting the ssh keys on local node which
        # otherwise cause ssh password prompts on the console (race conditions)
        # mode_delete() should be cleaning up the session tree
        if localdir and task == "delete":
            return

        pem_key_path = get_pem_key_path(args.session, args.volume)

        if not localdir:
            # prefix with ssh command if not local node
            cmd = ["ssh",
                   "-oNumberOfPasswordPrompts=0",
                   "-oStrictHostKeyChecking=no",
                   # We force TTY allocation (-t -t) so that Ctrl+C is handed
                   # through; see:
                   #   https://bugzilla.redhat.com/show_bug.cgi?id=1382236
                   # Note that this turns stderr of the remote `cmd`
                   # into stdout locally.
                   "-t",
                   "-t",
                   "-i", pem_key_path,
                   "root@%s" % host] + cmd

        (returncode, err, out) = execute(cmd, logger=logger)
        if returncode != 0:
            # Because the `-t -t` above turns the remote stderr into
            # local stdout, we need to log both stderr and stdout
            # here to print all error messages.
            fail("%s - %s failed; stdout (including remote stderr):\n"
                 "%s\n"
                 "stderr:\n"
                 "%s" % (host, task, out, err),
                 returncode,
                 logger=logger)

        if opts.get("copy_outfile", False) and not localdir:
            cmd_copy = ["scp",
                        "-oNumberOfPasswordPrompts=0",
                        "-oStrictHostKeyChecking=no",
                        "-i", pem_key_path,
                        "root@%s:/%s" % (host, opts.get("node_outfile")),
                        os.path.dirname(opts.get("node_outfile"))]
            execute(cmd_copy, exit_msg="%s - Copy command failed" % host,
                    logger=logger)
    except KeyboardInterrupt:
        sys.exit(2)


def run_cmd_nodes(task, args, **kwargs):
    global g_pid_nodefile_map
    nodes = get_nodes(args.volume)
    pool = []
    for num, node in enumerate(nodes):
        host, brick = node[1].split(":")
        host_uuid = node[0]
        cmd = []
        opts = {}

        # tmpfilename is valid only for tasks: pre, query and cleanup
        tmpfilename = kwargs.get("tmpfilename", "BADNAME")

        node_outfile = os.path.join(conf.get_opt("working_dir"),
                                    args.session, args.volume,
                                    tmpfilename,
                                    "tmp_output_%s" % num)

        if task == "pre":
            if vol_statusStr != "Started":
                fail("Volume %s is not online" % args.volume,
                     logger=logger)

            # If Full backup is requested or start time is zero, use brickfind
            change_detector = conf.get_change_detector("changelog")
            tag = None
            if args.full:
                change_detector = conf.get_change_detector("brickfind")
                tag = args.tag_for_full_find.strip()
                if tag == "":
                    tag = '""' if not is_host_local(host_uuid) else ""

            # remote file will be copied into this directory
            mkdirp(os.path.dirname(node_outfile),
                   exit_on_err=True, logger=logger)

            FS = args.field_separator
            if not is_host_local(host_uuid):
                FS = "'" + FS + "'"

            cmd = [change_detector,
                   args.session,
                   args.volume,
                   host,
                   brick,
                   node_outfile] + \
                ([str(kwargs.get("start")), str(kwargs.get("end"))]
                    if not args.full else []) + \
                ([tag] if tag is not None else []) + \
                ["--output-prefix", args.output_prefix] + \
                (["--debug"] if args.debug else []) + \
                (["--no-encode"] if args.no_encode else []) + \
                (["--only-namespace-changes"] if args.only_namespace_changes
                 else []) + \
                (["--type", args.type]) + \
                (["--field-separator", FS] if args.full else [])

            opts["node_outfile"] = node_outfile
            opts["copy_outfile"] = True
        elif task == "query":
            # If Full backup is requested or start time is zero, use brickfind
            tag = None
            change_detector = conf.get_change_detector("changelog")
            if args.full:
                change_detector = conf.get_change_detector("brickfind")
                tag = args.tag_for_full_find.strip()
                if tag == "":
                    tag = '""' if not is_host_local(host_uuid) else ""

            # remote file will be copied into this directory
            mkdirp(os.path.dirname(node_outfile),
                   exit_on_err=True, logger=logger)

            FS = args.field_separator
            if not is_host_local(host_uuid):
                FS = "'" + FS + "'"

            cmd = [change_detector,
                   args.session,
                   args.volume,
                   host,
                   brick,
                   node_outfile] + \
                ([str(kwargs.get("start")), str(kwargs.get("end"))]
                    if not args.full else []) + \
                ([tag] if tag is not None else []) + \
                ["--only-query"] + \
                ["--output-prefix", args.output_prefix] + \
                (["--debug"] if args.debug else []) + \
                (["--no-encode"] if args.no_encode else []) + \
                (["--only-namespace-changes"]
                    if args.only_namespace_changes else []) + \
                (["--type", args.type]) + \
                (["--field-separator", FS] if args.full else [])

            opts["node_outfile"] = node_outfile
            opts["copy_outfile"] = True
        elif task == "cleanup":
            # After pre/query run, cleanup the working directory and other
            # temp files. Remove the directory to which node_outfile has
            # been copied in main node
            try:
                os.remove(node_outfile)
            except (OSError, IOError):
                logger.warn("Failed to cleanup temporary file %s" %
                            node_outfile)
                pass

            cmd = [conf.get_opt("nodeagent"),
                   "cleanup",
                   args.session,
                   args.volume,
                   os.path.dirname(node_outfile)] + \
                (["--debug"] if args.debug else [])
        elif task == "create":
            if vol_statusStr != "Started":
                fail("Volume %s is not online" % args.volume,
                     logger=logger)

            # When glusterfind create, create session directory in
            # each brick nodes
            cmd = [conf.get_opt("nodeagent"),
                   "create",
                   args.session,
                   args.volume,
                   brick,
                   kwargs.get("time_to_update")] + \
                (["--debug"] if args.debug else []) + \
                (["--reset-session-time"] if args.reset_session_time
                 else [])
        elif task == "post":
            # Rename pre status file to actual status file in each node
            cmd = [conf.get_opt("nodeagent"),
                   "post",
                   args.session,
                   args.volume,
                   brick] + \
                (["--debug"] if args.debug else [])
        elif task == "delete":
            # When glusterfind delete, cleanup all the session files/dirs
            # from each node.
            cmd = [conf.get_opt("nodeagent"),
                   "delete",
                   args.session,
                   args.volume] + \
                (["--debug"] if args.debug else [])

        if cmd:
            p = Process(target=node_cmd,
                        args=(host, host_uuid, task, cmd, args, opts))
            p.start()
            pool.append(p)
            g_pid_nodefile_map[p.pid] = node_outfile

    for num, p in enumerate(pool):
        p.join()
        if p.exitcode != 0:
            logger.warn("Command %s failed in %s" % (task, nodes[num][1]))
            if task in ["create", "delete"]:
                fail("Command %s failed in %s" % (task, nodes[num][1]))
            elif task == "pre" or task == "query":
                if args.disable_partial:
                    sys.exit(1)
                else:
                    del g_pid_nodefile_map[p.pid]


@cache_output
def get_nodes(volume):
    """
    Get the gluster volume info xml output and parse to get
    the brick details.
    """
    global vol_statusStr

    cmd = ["gluster", 'volume', 'info', volume, "--xml"]
    _, data, _ = execute(cmd,
                         exit_msg="Failed to Run Gluster Volume Info",
                         logger=logger)
    tree = etree.fromstring(data)

    # Test to check if volume has been deleted after session creation
    count_el = tree.find('volInfo/volumes/count')
    if int(count_el.text) == 0:
        fail("Unable to get volume details", logger=logger)

    # this status is used in caller: run_cmd_nodes
    vol_statusStr = tree.find('volInfo/volumes/volume/statusStr').text
    vol_typeStr = tree.find('volInfo/volumes/volume/typeStr').text

    nodes = []
    volume_el = tree.find('volInfo/volumes/volume')
    try:
        brick_elems = []
        if vol_typeStr == "Tier":
            brick_elems.append('bricks/hotBricks/brick')
            brick_elems.append('bricks/coldBricks/brick')
        else:
            brick_elems.append('bricks/brick')

        for elem in brick_elems:
            for b in volume_el.findall(elem):
                nodes.append((b.find('hostUuid').text,
                              b.find('name').text))
    except (ParseError, AttributeError, ValueError) as e:
        fail("Failed to parse Volume Info: %s" % e, logger=logger)

    return nodes


def _get_args():
    parser = ArgumentParser(formatter_class=RawDescriptionHelpFormatter,
                            description=PROG_DESCRIPTION)
    subparsers = parser.add_subparsers(dest="mode")
    subparsers.required = True

    # create <SESSION> <VOLUME> [--debug] [--force]
    parser_create = subparsers.add_parser('create')
    parser_create.add_argument("session", help="Session Name")
    parser_create.add_argument("volume", help="Volume Name")
    parser_create.add_argument("--debug", help="Debug", action="store_true")
    parser_create.add_argument("--force", help="Force option to recreate "
                               "the session", action="store_true")
    parser_create.add_argument("--reset-session-time",
                               help="Reset Session Time to Current Time",
                               action="store_true")

    # delete <SESSION> <VOLUME> [--debug]
    parser_delete = subparsers.add_parser('delete')
    parser_delete.add_argument("session", help="Session Name")
    parser_delete.add_argument("volume", help="Volume Name")
    parser_delete.add_argument("--debug", help="Debug", action="store_true")

    # list [--session <SESSION>] [--volume <VOLUME>]
    parser_list = subparsers.add_parser('list')
    parser_list.add_argument("--session", help="Session Name", default="")
    parser_list.add_argument("--volume", help="Volume Name", default="")
    parser_list.add_argument("--debug", help="Debug", action="store_true")

    # pre <SESSION> <VOLUME> <OUTFILE>
    #     [--output-prefix <OUTPUT_PREFIX>] [--full]
    parser_pre = subparsers.add_parser('pre')
    parser_pre.add_argument("session", help="Session Name")
    parser_pre.add_argument("volume", help="Volume Name")
    parser_pre.add_argument("outfile", help="Output File", action=StoreAbsPath)
    parser_pre.add_argument("--debug", help="Debug", action="store_true")
    parser_pre.add_argument("--no-encode",
                            help="Do not encode path in output file",
                            action="store_true")
    parser_pre.add_argument("--full", help="Full find", action="store_true")
    parser_pre.add_argument("--disable-partial", help="Disable Partial find, "
                            "Fail when one node fails", action="store_true")
    parser_pre.add_argument("--output-prefix", help="File prefix in output",
                            default=".")
    parser_pre.add_argument("--regenerate-outfile",
                            help="Regenerate outfile, discard the outfile "
                            "generated from last pre command",
                            action="store_true")
    parser_pre.add_argument("-N", "--only-namespace-changes",
                            help="List only namespace changes",
                            action="store_true")
    parser_pre.add_argument("--tag-for-full-find",
                            help="Tag prefix for file names emitted during"
                            " a full find operation; default: \"NEW\"",
                            default="NEW")
    parser_pre.add_argument('--type', help="type: f, f-files only"
                            " d, d-directories only, by default = both",
                            default='both', choices=["f", "d", "both"])
    parser_pre.add_argument("--field-separator", help="Field separator string",
                            default=" ")

    # query <VOLUME> <OUTFILE> --since-time <SINCE_TIME>
    #       [--output-prefix <OUTPUT_PREFIX>] [--full]
    parser_query = subparsers.add_parser('query')
    parser_query.add_argument("volume", help="Volume Name")
    parser_query.add_argument("outfile", help="Output File",
                              action=StoreAbsPath)
    parser_query.add_argument("--since-time", help="UNIX epoch time since "
                              "which listing is required", type=int)
    parser_query.add_argument("--end-time", help="UNIX epoch time up to "
                              "which listing is required", type=int)
    parser_query.add_argument("--no-encode",
                              help="Do not encode path in output file",
                              action="store_true")
    parser_query.add_argument("--full", help="Full find", action="store_true")
    parser_query.add_argument("--debug", help="Debug", action="store_true")
    parser_query.add_argument("--disable-partial", help="Disable Partial find,"
                              " Fail when one node fails", action="store_true")
    parser_query.add_argument("--output-prefix", help="File prefix in output",
                              default=".")
    parser_query.add_argument("-N", "--only-namespace-changes",
                              help="List only namespace changes",
                              action="store_true")
    parser_query.add_argument("--tag-for-full-find",
                              help="Tag prefix for file names emitted during"
                              " a full find operation; default: \"NEW\"",
                              default="NEW")
    parser_query.add_argument('--type', help="type: f, f-files only"
                              " d, d-directories only, by default = both",
                              default='both', choices=["f", "d", "both"])
    parser_query.add_argument("--field-separator",
                              help="Field separator string",
                              default=" ")

    # post <SESSION> <VOLUME>
    parser_post = subparsers.add_parser('post')
    parser_post.add_argument("session", help="Session Name")
    parser_post.add_argument("volume", help="Volume Name")
    parser_post.add_argument("--debug", help="Debug", action="store_true")

    return parser.parse_args()


def ssh_setup(args):
    pem_key_path = get_pem_key_path(args.session, args.volume)

    if not os.path.exists(pem_key_path):
        # Generate ssh-key
        cmd = ["ssh-keygen",
               "-N",
               "",
               "-f",
               pem_key_path]
        execute(cmd,
                exit_msg="Unable to generate ssh key %s"
                % pem_key_path,
                logger=logger)

        logger.info("Ssh key generated %s" % pem_key_path)

    try:
        shutil.copyfile(pem_key_path + ".pub",
                        os.path.join(conf.get_opt("session_dir"),
                                     ".keys",
                                     "%s_%s_secret.pem.pub" % (args.session,
                                                               args.volume)))
    except (IOError, OSError) as e:
        fail("Failed to copy public key to %s: %s"
             % (os.path.join(conf.get_opt("session_dir"), ".keys"), e),
             logger=logger)

    # Copy pub file to all nodes
    cmd = ["gluster",
           "system::",
           "copy",
           "file",
           "/glusterfind/.keys/%s.pub" % os.path.basename(pem_key_path)]

    execute(cmd, exit_msg="Failed to distribute ssh keys", logger=logger)

    logger.info("Distributed ssh key to all nodes of Volume")

    # Add to authorized_keys file in each node
    cmd = ["gluster",
           "system::",
           "execute",
           "add_secret_pub",
           "root",
           "/glusterfind/.keys/%s.pub" % os.path.basename(pem_key_path)]
    execute(cmd,
            exit_msg="Failed to add ssh keys to authorized_keys file",
            logger=logger)

    logger.info("Ssh key added to authorized_keys of Volume nodes")


def enable_volume_options(args):
    execute(["gluster", "volume", "set",
             args.volume, "build-pgfid", "on"],
            exit_msg="Failed to set volume option build-pgfid on",
            logger=logger)
    logger.info("Volume option set %s, build-pgfid on" % args.volume)

    execute(["gluster", "volume", "set",
             args.volume, "changelog.changelog", "on"],
            exit_msg="Failed to set volume option "
            "changelog.changelog on", logger=logger)
    logger.info("Volume option set %s, changelog.changelog on"
                % args.volume)

    execute(["gluster", "volume", "set",
             args.volume, "changelog.capture-del-path", "on"],
            exit_msg="Failed to set volume option "
            "changelog.capture-del-path on", logger=logger)
    logger.info("Volume option set %s, changelog.capture-del-path on"
                % args.volume)


def write_output(outfile, outfilemerger, field_separator):
    with codecs.open(outfile, "a", encoding="utf-8") as f:
        for row in outfilemerger.get():
            # Multiple paths in case of Hardlinks
            paths = row[1].split(",")
            row_2_rep = None
            for p in paths:
                if p == "":
                    continue
                p_rep = p.replace("//", "/")
                if not row_2_rep:
                    row_2_rep = row[2].replace("//", "/")
                if p_rep == row_2_rep:
                    continue

                if row_2_rep and row_2_rep != "":
                    gfind_write_row(f, row[0], field_separator, p_rep, row_2_rep)

                else:
                    gfind_write(f, row[0], field_separator, p_rep)

def validate_volume(volume):
    cmd = ["gluster", 'volume', 'info', volume, "--xml"]
    _, data, _ = execute(cmd,
                         exit_msg="Failed to Run Gluster Volume Info",
                         logger=logger)
    try:
        tree = etree.fromstring(data)
        statusStr = tree.find('volInfo/volumes/volume/statusStr').text
    except (ParseError, AttributeError) as e:
        fail("Invalid Volume: Check the Volume name! %s" % e)
    if statusStr != "Started":
        fail("Volume %s is not online" % volume)

# The rules for a valid session name.
SESSION_NAME_RULES = {
    'min_length': 2,
    'max_length': 256,  # same as maximum volume length
    # Specifies all alphanumeric characters, underscore, hyphen.
    'valid_chars': r'0-9a-zA-Z_-',
}


# checks valid session name, fail otherwise
def validate_session_name(session):
    # Check for minimum length
    if len(session) < SESSION_NAME_RULES['min_length']:
        fail('session_name must be at least ' +
                 str(SESSION_NAME_RULES['min_length']) + ' characters long.')
    # Check for maximum length
    if len(session) > SESSION_NAME_RULES['max_length']:
        fail('session_name must not exceed ' +
                 str(SESSION_NAME_RULES['max_length']) + ' characters length.')

    # Matches strings composed entirely of characters specified within
    if not re.match(r'^[' + SESSION_NAME_RULES['valid_chars'] +
                        ']+$', session):
        fail('Session name can only contain these characters: ' +
                         SESSION_NAME_RULES['valid_chars'])


def mode_create(session_dir, args):
    validate_session_name(args.session)

    logger.debug("Init is called - Session: %s, Volume: %s"
                 % (args.session, args.volume))
    mkdirp(session_dir, exit_on_err=True, logger=logger)
    mkdirp(os.path.join(session_dir, args.volume), exit_on_err=True,
           logger=logger)
    status_file = os.path.join(session_dir, args.volume, "status")

    if os.path.exists(status_file) and not args.force:
        fail("Session %s already created" % args.session, logger=logger)

    if not os.path.exists(status_file) or args.force:
        ssh_setup(args)
        enable_volume_options(args)

    # Add Rollover time to current time to make sure changelogs
    # will be available if we use this time as start time
    time_to_update = int(time.time()) + get_changelog_rollover_time(
        args.volume)

    run_cmd_nodes("create", args, time_to_update=str(time_to_update))

    if not os.path.exists(status_file) or args.reset_session_time:
        with open(status_file, "w") as f:
            f.write(str(time_to_update))

    sys.stdout.write("Session %s created with volume %s\n" %
                     (args.session, args.volume))

    sys.exit(0)


def mode_query(session_dir, args):
    global gtmpfilename
    global g_pid_nodefile_map

    # Verify volume status
    cmd = ["gluster", 'volume', 'info', args.volume, "--xml"]
    _, data, _ = execute(cmd,
                         exit_msg="Failed to Run Gluster Volume Info",
                         logger=logger)
    try:
        tree = etree.fromstring(data)
        statusStr = tree.find('volInfo/volumes/volume/statusStr').text
    except (ParseError, AttributeError) as e:
        fail("Invalid Volume: %s" % e, logger=logger)

    if statusStr != "Started":
        fail("Volume %s is not online" % args.volume, logger=logger)

    mkdirp(session_dir, exit_on_err=True, logger=logger)
    mkdirp(os.path.join(session_dir, args.volume), exit_on_err=True,
           logger=logger)
    mkdirp(os.path.dirname(args.outfile), exit_on_err=True, logger=logger)

    # Configure cluster for pasword-less SSH
    ssh_setup(args)

    # Enable volume options for changelog capture
    enable_volume_options(args)

    # Test options
    if not args.full and args.type in ["f", "d"]:
        fail("--type can only be used with --full")
    if not args.since_time and not args.end_time and not args.full:
        fail("Please specify either {--since-time and optionally --end-time} "
             "or --full", logger=logger)

    if args.since_time and args.end_time and args.full:
        fail("Please specify either {--since-time and optionally --end-time} "
             "or --full, but not both",
             logger=logger)

    if args.end_time and not args.since_time:
        fail("Please specify --since-time as well", logger=logger)

    # Start query command processing
    start = -1
    end = -1
    if args.since_time:
        start = args.since_time
        if args.end_time:
            end = args.end_time
    else:
        start = 0  # --full option is handled separately

    logger.debug("Query is called - Session: %s, Volume: %s, "
                 "Start time: %s, End time: %s"
                 % ("default", args.volume, start, end))

    prefix = datetime.now().strftime("%Y%m%d-%H%M%S-%f-")
    gtmpfilename = prefix + next(tempfile._get_candidate_names())

    run_cmd_nodes("query", args, start=start, end=end,
                  tmpfilename=gtmpfilename)

    # Merger
    if args.full:
        if len(g_pid_nodefile_map) > 0:
            cmd = ["sort", "-u"] + list(g_pid_nodefile_map.values()) + \
                  ["-o", args.outfile]
            execute(cmd,
                    exit_msg="Failed to merge output files "
                    "collected from nodes", logger=logger)
        else:
            fail("Failed to collect any output files from peers. "
                 "Looks like all bricks are offline.", logger=logger)
    else:
        # Read each Changelogs db and generate finaldb
        create_file(args.outfile, exit_on_err=True, logger=logger)
        outfilemerger = OutputMerger(args.outfile + ".db",
                                     list(g_pid_nodefile_map.values()))
        write_output(args.outfile, outfilemerger, args.field_separator)

    try:
        os.remove(args.outfile + ".db")
    except (IOError, OSError):
        pass

    run_cmd_nodes("cleanup", args, tmpfilename=gtmpfilename)

    sys.stdout.write("Generated output file %s\n" % args.outfile)


def mode_pre(session_dir, args):
    global gtmpfilename
    global g_pid_nodefile_map

    """
    Read from Session file and write to session.pre file
    """
    endtime_to_update = int(time.time()) - get_changelog_rollover_time(
        args.volume)
    status_file = os.path.join(session_dir, args.volume, "status")
    status_file_pre = status_file + ".pre"

    mkdirp(os.path.dirname(args.outfile), exit_on_err=True, logger=logger)

    if not args.full and args.type in ["f", "d"]:
        fail("--type can only be used with --full")

    # If Pre status file exists and running pre command again
    if os.path.exists(status_file_pre) and not args.regenerate_outfile:
        fail("Post command is not run after last pre, "
             "use --regenerate-outfile")

    start = 0
    try:
        with open(status_file) as f:
            start = int(f.read().strip())
    except ValueError:
        pass
    except (OSError, IOError) as e:
        fail("Error Opening Session file %s: %s"
             % (status_file, e), logger=logger)

    logger.debug("Pre is called - Session: %s, Volume: %s, "
                 "Start time: %s, End time: %s"
                 % (args.session, args.volume, start, endtime_to_update))

    prefix = datetime.now().strftime("%Y%m%d-%H%M%S-%f-")
    gtmpfilename = prefix + next(tempfile._get_candidate_names())

    run_cmd_nodes("pre", args, start=start, end=-1, tmpfilename=gtmpfilename)

    # Merger
    if args.full:
        if len(g_pid_nodefile_map) > 0:
            cmd = ["sort", "-u"] + list(g_pid_nodefile_map.values()) + \
                  ["-o", args.outfile]
            execute(cmd,
                    exit_msg="Failed to merge output files "
                    "collected from nodes", logger=logger)
        else:
            fail("Failed to collect any output files from peers. "
                 "Looks like all bricks are offline.", logger=logger)
    else:
        # Read each Changelogs db and generate finaldb
        create_file(args.outfile, exit_on_err=True, logger=logger)
        outfilemerger = OutputMerger(args.outfile + ".db",
                                     list(g_pid_nodefile_map.values()))
        write_output(args.outfile, outfilemerger, args.field_separator)

    try:
        os.remove(args.outfile + ".db")
    except (IOError, OSError):
        pass

    run_cmd_nodes("cleanup", args, tmpfilename=gtmpfilename)

    with open(status_file_pre, "w") as f:
        f.write(str(endtime_to_update))

    sys.stdout.write("Generated output file %s\n" % args.outfile)


def mode_post(session_dir, args):
    """
    If pre session file exists, overwrite session file
    If pre session file does not exists, return ERROR
    """
    status_file = os.path.join(session_dir, args.volume, "status")
    logger.debug("Post is called - Session: %s, Volume: %s"
                 % (args.session, args.volume))
    status_file_pre = status_file + ".pre"

    if os.path.exists(status_file_pre):
        run_cmd_nodes("post", args)
        os.rename(status_file_pre, status_file)
        sys.stdout.write("Session %s with volume %s updated\n" %
                         (args.session, args.volume))
        sys.exit(0)
    else:
        fail("Pre script is not run", logger=logger)


def mode_delete(session_dir, args):
    run_cmd_nodes("delete", args)
    shutil.rmtree(os.path.join(session_dir, args.volume),
                  onerror=handle_rm_error)
    sys.stdout.write("Session %s with volume %s deleted\n" %
                     (args.session, args.volume))

    # If the session contains only this volume, then cleanup the
    # session directory. If a session contains multiple volumes
    # then os.rmdir will fail with ENOTEMPTY
    try:
        os.rmdir(session_dir)
    except OSError as e:
        if not e.errno == ENOTEMPTY:
            logger.warn("Failed to delete session directory: %s" % e)


def mode_list(session_dir, args):
    """
    List available sessions to stdout, if session name is set
    only list that session.
    """
    if args.session:
        if not os.path.exists(os.path.join(session_dir, args.session)):
            fail("Invalid Session", logger=logger)
        sessions = [args.session]
    else:
        sessions = []
        for d in os.listdir(session_dir):
            if d != ".keys":
                sessions.append(d)

    output = []
    for session in sessions:
        # Session Volume Last Processed
        volnames = os.listdir(os.path.join(session_dir, session))

        for volname in volnames:
            if args.volume and args.volume != volname:
                continue

            status_file = os.path.join(session_dir, session, volname, "status")
            last_processed = None
            try:
                with open(status_file) as f:
                    last_processed = f.read().strip()
            except (OSError, IOError) as e:
                if e.errno == ENOENT:
                    continue
                else:
                    raise
            output.append((session, volname, last_processed))

    if output:
        sys.stdout.write("%s %s %s\n" % ("SESSION".ljust(25),
                                         "VOLUME".ljust(25),
                                         "SESSION TIME".ljust(25)))
        sys.stdout.write("-"*75)
        sys.stdout.write("\n")
    for session, volname, last_processed in output:
        sess_time = 'Session Corrupted'
        if last_processed:
            try:
                sess_time = human_time(last_processed)
            except TypeError:
                sess_time = 'Session Corrupted'
        sys.stdout.write("%s %s %s\n" % (session.ljust(25),
                                         volname.ljust(25),
                                         sess_time.ljust(25)))

    if not output:
        if args.session or args.volume:
            fail("Invalid Session", logger=logger)
        else:
            sys.stdout.write("No sessions found.\n")


def main():
    global gtmpfilename

    args = None

    try:
        args = _get_args()
        mkdirp(conf.get_opt("session_dir"), exit_on_err=True)

        # force the default session name if mode is "query"
        if args.mode == "query":
            args.session = "default"

        if args.mode == "list":
            session_dir = conf.get_opt("session_dir")
        else:
            session_dir = os.path.join(conf.get_opt("session_dir"),
                                       args.session)

        if not os.path.exists(session_dir) and \
                args.mode not in ["create", "list", "query"]:
            fail("Invalid session %s" % args.session)

        # volume involved, validate the volume first
        if args.mode not in ["list"]:
            validate_volume(args.volume)


        # "default" is a system defined session name
        if args.mode in ["create", "post", "pre", "delete"] and \
                args.session == "default":
            fail("Invalid session %s" % args.session)

        vol_dir = os.path.join(session_dir, args.volume)
        if not os.path.exists(vol_dir) and args.mode not in \
                ["create", "list", "query"]:
            fail("Session %s not created with volume %s" %
                 (args.session, args.volume))

        mkdirp(os.path.join(conf.get_opt("log_dir"),
                            args.session,
                            args.volume),
               exit_on_err=True)
        log_file = os.path.join(conf.get_opt("log_dir"),
                                args.session,
                                args.volume,
                                "cli.log")
        setup_logger(logger, log_file, args.debug)

        # globals() will have all the functions already defined.
        # mode_<args.mode> will be the function name to be called
        globals()["mode_" + args.mode](session_dir, args)
    except KeyboardInterrupt:
        if args is not None:
            if args.mode == "pre" or args.mode == "query":
                # cleanup session
                if gtmpfilename is not None:
                    # no more interrupts until we clean up
                    signal.signal(signal.SIGINT, signal.SIG_IGN)
                    run_cmd_nodes("cleanup", args, tmpfilename=gtmpfilename)

        # Interrupted, exit with non zero error code
        sys.exit(2)



# -*- coding: utf-8 -*-
#
#  Copyright (c) 2016 Red Hat, Inc. <http://www.redhat.com>
#  This file is part of GlusterFS.
#
#  This file is licensed to you under your choice of the GNU Lesser
#  General Public License, version 3 or any later version (LGPLv3 or
#  later), or the GNU General Public License, version 2 (GPLv2), in all
#  cases as published by the Free Software Foundation.
#

import utils


def generic_handler(ts, key, data):
    """
    Generic handler to broadcast message to all peers, custom handlers
    can be created by func name handler_<event_name>
    Ex: handle_event_volume_create(ts, key, data)
    """
    utils.publish(ts, key, data)


def handle_event_volume_set(ts, key, data):
    """
    Received data will have all the options as one string, split into
    list of options. "key1,value1,key2,value2" into
    [[key1, value1], [key2, value2]]
    """
    opts = data.get("options", "").strip(",").split(",")
    data["options"] = []
    for i, opt in enumerate(opts):
        if i % 2 == 0:
            # Add new array with key
            data["options"].append([opt])
        else:
            # Add to the last added array
            data["options"][-1].append(opt)

    utils.publish(ts, key, data)



#!/usr/bin/python3
"""
Schedule Geo-replication
------------------------
A tool to run Geo-replication when required. This can be used to
schedule the Geo-replication to run once in a day using

   # Run daily at 08:30pm
   30 20 * * * root python /usr/share/glusterfs/scripts/schedule_georep.py \\
      --no-color gv1 fvm1 gv2 >> /var/log/glusterfs/schedule_georep.log 2>&1

This tool does the following,

1. Stop Geo-replication if Started
2. Start Geo-replication
3. Set Checkpoint
4. Check the Status and see Checkpoint is Complete.(LOOP)
5. If checkpoint complete, Stop Geo-replication

Usage:

    python /usr/share/glusterfs/scripts/schedule_georep.py <PRIMARYVOL> \\
         <SECONDARYHOST> <SECONDARYVOL>

For example,

    python /usr/share/glusterfs/scripts/schedule_georep.py gv1 fvm1 gv2

"""
import subprocess
import time
import xml.etree.cElementTree as etree
import sys
from contextlib import contextmanager
import tempfile
import os
from argparse import ArgumentParser, RawDescriptionHelpFormatter

ParseError = etree.ParseError if hasattr(etree, 'ParseError') else SyntaxError
cache_data = {}

SESSION_MOUNT_LOG_FILE = ("/var/log/glusterfs/geo-replication"
                          "/schedule_georep.mount.log")

USE_CLI_COLOR = True
mnt_list = []

class GlusterBadXmlFormat(Exception):
    """
    Exception class for XML Parse Errors
    """
    pass


def output_notok(msg, err="", exitcode=1):
    if USE_CLI_COLOR:
        out = "\033[31m[NOT OK]\033[0m {0}\n{1}\n"
    else:
        out = "[NOT OK] {0}\n{1}\n"
    sys.stderr.write(out.format(msg, err))
    sys.exit(exitcode)


def output_warning(msg):
    if USE_CLI_COLOR:
        out = "\033[33m[  WARN]\033[0m {0}\n"
    else:
        out = "[  WARN] {0}\n"
    sys.stderr.write(out.format(msg))


def output_ok(msg):
    if USE_CLI_COLOR:
        out = "\033[32m[    OK]\033[0m {0}\n"
    else:
        out = "[    OK] {0}\n"
    sys.stderr.write(out.format(msg))


def execute(cmd, success_msg="", failure_msg="", exitcode=-1):
    """
    Generic wrapper to execute the CLI commands. Returns Output if success.
    On success it can print message in stdout if specified.
    On failure, exits after writing to stderr.
    """
    p = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, universal_newlines=True)
    out, err = p.communicate()
    if p.returncode == 0:
        if success_msg:
            output_ok(success_msg)
        return out
    else:
        if exitcode == 0:
            return
        err_msg = err if err else out
        output_notok(failure_msg, err=err_msg, exitcode=exitcode)


def cache_output_with_args(func):
    """
    Decorator function to remember the output of any function
    """
    def wrapper(*args, **kwargs):
        global cache_data
        key = "_".join([func.__name__] + list(args))
        if cache_data.get(key, None) is None:
            cache_data[key] = func(*args, **kwargs)

        return cache_data[key]
    return wrapper


def cleanup(hostname, volname, mnt):
    """
    Unmount the Volume and Remove the temporary directory
    """
    execute(["umount", "-l", mnt],
            failure_msg="Unable to Unmount Gluster Volume "
            "{0}:{1}(Mounted at {2})".format(hostname, volname, mnt))
    execute(["rmdir", mnt],
            failure_msg="Unable to Remove temp directory "
            "{0}".format(mnt), exitcode=0)


@contextmanager
def glustermount(hostname, volname):
    """
    Context manager for Mounting Gluster Volume
    Use as
        with glustermount(HOSTNAME, VOLNAME) as MNT:
            # Do your stuff
    Automatically unmounts it in case of Exceptions/out of context
    """
    mnt = tempfile.mkdtemp(prefix="georepsetup_")
    mnt_list.append(mnt)
    execute(["/usr/bin/glusterfs",
             "--volfile-server", hostname,
             "--volfile-id", volname,
             "-l", SESSION_MOUNT_LOG_FILE,
             mnt],
            failure_msg="Unable to Mount Gluster Volume "
            "{0}:{1}".format(hostname, volname))
    if os.path.ismount(mnt):
        yield mnt
    else:
        output_notok("Unable to Mount Gluster Volume "
                     "{0}:{1}".format(hostname, volname))
    cleanup(hostname, volname, mnt)


@cache_output_with_args
def get_bricks(volname):
    """
    Returns Bricks list, caches the Bricks list for a volume once
    parsed.
    """
    value = []
    cmd = ["/usr/bin/gluster", "volume", "info", volname, "--xml"]
    info = execute(cmd)
    try:
        tree = etree.fromstring(info)
        volume_el = tree.find('volInfo/volumes/volume')
        for b in volume_el.findall('bricks/brick'):
            value.append({"name": b.find("name").text,
                          "hostUuid": b.find("hostUuid").text})
    except ParseError:
        raise GlusterBadXmlFormat("Bad XML Format: %s" % " ".join(cmd))

    return value


def get_georep_status(primaryvol, secondary):
    session_keys = set()
    out = {}
    cmd = ["/usr/bin/gluster", "volume", "geo-replication"]
    if primaryvol is not None:
        cmd += [primaryvol]
        if secondary:
            cmd += [secondary]

    cmd += ["status", "--xml"]
    info = execute(cmd)

    try:
        tree = etree.fromstring(info)
        # Get All Sessions
        for volume_el in tree.findall("geoRep/volume"):
            sessions_el = volume_el.find("sessions")
            # Primary Volume name if multiple Volumes
            mvol = volume_el.find("name").text

            # For each session, collect the details
            for session in sessions_el.findall("session"):
                session_secondary = "{0}:{1}".format(mvol, session.find(
                    "session_secondary").text)
                session_keys.add(session_secondary)
                out[session_secondary] = {}

                for pair in session.findall('pair'):
                    primary_brick = "{0}:{1}".format(
                        pair.find("primary_node").text,
                        pair.find("primary_brick").text
                    )

                    out[session_secondary][primary_brick] = {
                        "primaryvol": mvol,
                        "secondaryvol": pair.find("secondary").text.split("::")[-1],
                        "primary_node": pair.find("primary_node").text,
                        "primary_brick": pair.find("primary_brick").text,
                        "secondary_user": pair.find("secondary_user").text,
                        "secondary": pair.find("secondary").text,
                        "secondary_node": pair.find("secondary_node").text,
                        "status": pair.find("status").text,
                        "crawl_status": pair.find("crawl_status").text,
                        "entry": pair.find("entry").text,
                        "data": pair.find("data").text,
                        "meta": pair.find("meta").text,
                        "failures": pair.find("failures").text,
                        "checkpoint_completed": pair.find(
                            "checkpoint_completed").text,
                        "primary_node_uuid": pair.find("primary_node_uuid").text,
                        "last_synced": pair.find("last_synced").text,
                        "checkpoint_time": pair.find("checkpoint_time").text,
                        "checkpoint_completion_time":
                        pair.find("checkpoint_completion_time").text
                    }
    except ParseError:
        raise GlusterBadXmlFormat("Bad XML Format: %s" % " ".join(cmd))

    return session_keys, out


def get_offline_status(volname, brick, node_uuid, secondary):
    node, brick = brick.split(":")
    if "@" not in secondary:
        secondary_user = "root"
    else:
        secondary_user, _ = secondary.split("@")

    return {
        "primaryvol": volname,
        "secondaryvol": secondary.split("::")[-1],
        "primary_node": node,
        "primary_brick": brick,
        "secondary_user": secondary_user,
        "secondary": secondary,
        "secondary_node": "N/A",
        "status": "Offline",
        "crawl_status": "N/A",
        "entry": "N/A",
        "data": "N/A",
        "meta": "N/A",
        "failures": "N/A",
        "checkpoint_completed": "N/A",
        "primary_node_uuid": node_uuid,
        "last_synced": "N/A",
        "checkpoint_time": "N/A",
        "checkpoint_completion_time": "N/A"
    }


def get(primaryvol=None, secondary=None):
    """
    This function gets list of Bricks of Primary Volume and collects
    respective Geo-rep status. Output will be always ordered as the
    bricks list in Primary Volume. If Geo-rep status is not available
    for any brick then it updates OFFLINE status.
    """
    out = []
    session_keys, gstatus = get_georep_status(primaryvol, secondary)

    for session in session_keys:
        mvol, _, secondary = session.split(":", 2)
        secondary = secondary.replace("ssh://", "")
        primary_bricks = get_bricks(mvol)
        out.append([])
        for brick in primary_bricks:
            bname = brick["name"]
            if gstatus.get(session) and gstatus[session].get(bname, None):
                out[-1].append(gstatus[session][bname])
            else:
                out[-1].append(
                    get_offline_status(mvol, bname, brick["hostUuid"], secondary))

    return out


def get_summary(primaryvol, secondary_url):
    """
    Wrapper function around Geo-rep Status and Gluster Volume Info
    This combines the output from Bricks list and Geo-rep Status.
    If a Primary Brick node is down or Status is faulty then increments
    the faulty counter. It also collects the checkpoint status from all
    workers and compares with Number of Bricks.
    """
    down_rows = []
    faulty_rows = []
    out = []

    status_data = get(primaryvol, secondary_url)

    for session in status_data:
        session_name = ""
        summary = {
            "active": 0,
            "passive": 0,
            "faulty": 0,
            "initializing": 0,
            "stopped": 0,
            "created": 0,
            "offline": 0,
            "paused": 0,
            "workers": 0,
            "completed_checkpoints": 0,
            "checkpoint": False,
            "checkpoints_ok": False,
            "ok": False
        }

        for row in session:
            summary[row["status"].replace("...", "").lower()] += 1
            summary["workers"] += 1
            if row["checkpoint_completed"] == "Yes":
                summary["completed_checkpoints"] += 1

            session_name = "{0}=>{1}".format(
                row["primaryvol"],
                row["secondary"].replace("ssh://", "")
            )

            if row["status"] == "Faulty":
                faulty_rows.append("{0}:{1}".format(row["primary_node"],
                                                    row["primary_brick"]))

            if row["status"] == "Offline":
                down_rows.append("{0}:{1}".format(row["primary_node"],
                                                  row["primary_brick"]))

        if summary["active"] == summary["completed_checkpoints"] and \
           summary["faulty"] == 0 and summary["offline"] == 0:
            summary["checkpoints_ok"] = True

        if summary["faulty"] == 0 and summary["offline"] == 0:
            summary["ok"] = True

        if session_name != "":
                out.append([session_name, summary, faulty_rows, down_rows])

    return out


def touch_mount_root(primaryvol):
    # Create a Mount and Touch the Mount point root,
    # Hack to make sure some event available after
    # setting Checkpoint. Without this there is a chance of
    # Checkpoint never completes.
    with glustermount("localhost", primaryvol) as mnt:
        execute(["touch", mnt])


def main(args):
    turns = 1

    # Stop Force
    cmd = ["/usr/bin/gluster", "volume", "geo-replication", args.primaryvol,
           "%s::%s" % (args.secondary, args.secondaryvol), "stop", "force"]
    execute(cmd)
    output_ok("Stopped Geo-replication")

    # Set Checkpoint to NOW
    cmd = ["/usr/bin/gluster", "volume", "geo-replication", args.primaryvol,
           "%s::%s" % (args.secondary, args.secondaryvol), "config", "checkpoint",
           "now"]
    execute(cmd)
    output_ok("Set Checkpoint")

    # Start the Geo-replication
    cmd = ["/usr/bin/gluster", "volume", "geo-replication", args.primaryvol,
           "%s::%s" % (args.secondary, args.secondaryvol), "start"]
    execute(cmd)
    output_ok("Started Geo-replication and watching Status for "
              "Checkpoint completion")

    start_time = int(time.time())
    duration = 0

    # Sleep till Geo-rep initializes
    time.sleep(60)

    touch_mount_root(args.primaryvol)

    secondary_url = "{0}::{1}".format(args.secondary, args.secondaryvol)

    # Loop to Check the Geo-replication Status and Checkpoint
    # If All Status OK and all Checkpoints complete,
    # Stop the Geo-replication and Log the Completeness
    while True:
        session_summary = get_summary(args.primaryvol,
                                      secondary_url)
        if len(session_summary) == 0:
            # If Status command fails with another transaction error
            # or any other error. Gluster cmd still produces XML output
            # with different message
            output_warning("Unable to get Geo-replication Status")
        else:
            session_name, summary, faulty_rows, down_rows = session_summary[0]
            chkpt_status = "COMPLETE" if summary["checkpoints_ok"] else \
                           "NOT COMPLETE"
            ok_status = "OK" if summary["ok"] else "NOT OK"

            if summary["ok"]:
                output_ok("All Checkpoints {1}, "
                          "All status {2} (Turns {0:>3})".format(
                              turns, chkpt_status, ok_status))
            else:
                output_warning("All Checkpoints {1}, "
                               "All status {2} (Turns {0:>3})".format(
                                   turns, chkpt_status, ok_status))

                output_warning("Geo-rep workers Faulty/Offline, "
                               "Faulty: {0} Offline: {1}".format(
                                   repr(faulty_rows),
                                   repr(down_rows)))

            if summary["checkpoints_ok"]:
                output_ok("Stopping Geo-replication session now")
                cmd = ["/usr/bin/gluster", "volume", "geo-replication",
                       args.primaryvol,
                       "%s::%s" % (args.secondary, args.secondaryvol), "stop"]
                execute(cmd)
                break
            else:
                # If Checkpoint is not complete after a iteration means brick
                # was down and came online now. SETATTR on mount is not
                # recorded, So again issue touch on mount root So that
                # Stime will increase and Checkpoint will complete.
                touch_mount_root(args.primaryvol)

        # Increment the turns and Sleep for 10 sec
        turns += 1
        duration = int(time.time()) - start_time
        if args.timeout > 0 and duration > (args.timeout * 60):
            cmd = ["/usr/bin/gluster", "volume", "geo-replication",
                   args.primaryvol,
                   "%s::%s" % (args.secondary, args.secondaryvol), "stop", "force"]
            execute(cmd)
            output_notok("Timed out, Stopping Geo-replication("
                         "Duration: {0}sec)".format(duration))

        time.sleep(args.interval)

    for mnt in mnt_list:
        execute(["rmdir", mnt],
                failure_msg="Unable to Remove temp directory "
                "{0}".format(mnt), exitcode=0)

if __name__ == "__main__":
    parser = ArgumentParser(formatter_class=RawDescriptionHelpFormatter,
                            description=__doc__)
    parser.add_argument("primaryvol", help="Primary Volume Name")
    parser.add_argument("secondary",
                        help="Secondary hostname "
                        "(<username>@SECONDARYHOST or SECONDARYHOST)",
                        metavar="SECONDARY")
    parser.add_argument("secondaryvol", help="Secondary Volume Name")
    parser.add_argument("--interval", help="Interval in Seconds. "
                        "Wait time before each status check",
                        type=int, default=10)
    parser.add_argument("--timeout", help="Timeout in minutes. Script will "
                        "stop Geo-replication if Checkpoint is not complete "
                        "in the specified timeout time", type=int,
                        default=0)
    parser.add_argument("--no-color", help="Don't use Color in CLI output",
                        action="store_true")
    args = parser.parse_args()
    if args.no_color:
        USE_CLI_COLOR = False
    try:
        # Check for session existence
        cmd = ["/usr/bin/gluster", "volume", "geo-replication",
               args.primaryvol, "%s::%s" % (args.secondary, args.secondaryvol), "status"]
        execute(cmd)
        main(args)
    except KeyboardInterrupt:
        for mnt in mnt_list:
            execute(["umount", "-l", mnt],
                    failure_msg="Unable to Unmount Gluster Volume "
                    "Mounted at {0}".format(mnt), exitcode=0)
            execute(["rmdir", mnt],
                    failure_msg="Unable to Remove temp directory "
                    "{0}".format(mnt), exitcode=0)
        output_notok("Exiting...")



#
# Copyright (c) 2011-2014 Red Hat, Inc. <http://www.redhat.com>
# This file is part of GlusterFS.

# This file is licensed to you under your choice of the GNU Lesser
# General Public License, version 3 or any later version (LGPLv3 or
# later), or the GNU General Public License, version 2 (GPLv2), in all
# cases as published by the Free Software Foundation.
#

import os
from ctypes import CDLL, get_errno
from py2py3 import (bytearray_to_str, gr_create_string_buffer,
                    gr_query_xattr, gr_lsetxattr, gr_lremovexattr)


class Xattr(object):

    """singleton that wraps the extended attributes system
       interface for python using ctypes

       Just implement it to the degree we need it, in particular
       - we need just the l*xattr variants, ie. we never want symlinks to be
         followed
       - don't need size discovery for getxattr, as we always know the exact
         sizes we expect
    """

    libc = CDLL("libc.so.6", use_errno=True)

    @classmethod
    def geterrno(cls):
        return get_errno()

    @classmethod
    def raise_oserr(cls):
        errn = cls.geterrno()
        raise OSError(errn, os.strerror(errn))

    @classmethod
    def _query_xattr(cls, path, siz, syscall, *a):
        if siz:
            buf = gr_create_string_buffer(siz)
        else:
            buf = None
        ret = getattr(cls.libc, syscall)(*((path,) + a + (buf, siz)))
        if ret == -1:
            cls.raise_oserr()
        if siz:
            # py2 and py3 compatibility. Convert bytes array
            # to string
            result = bytearray_to_str(buf.raw)
            return result[:ret]
        else:
            return ret

    @classmethod
    def lgetxattr(cls, path, attr, siz=0):
        return gr_query_xattr(cls, path, siz, 'lgetxattr', attr)

    @classmethod
    def lgetxattr_buf(cls, path, attr):
        """lgetxattr variant with size discovery"""
        size = cls.lgetxattr(path, attr)
        if size == -1:
            cls.raise_oserr()
        if size == 0:
            return ''
        return cls.lgetxattr(path, attr, size)

    @classmethod
    def llistxattr(cls, path, siz=0):
        ret = gr_query_xattr(cls, path, siz, 'llistxattr')
        if isinstance(ret, str):
            ret = ret.strip('\0')
            ret = ret.split('\0') if ret else []
        return ret

    @classmethod
    def lsetxattr(cls, path, attr, val):
        ret = gr_lsetxattr(cls, path, attr, val)
        if ret == -1:
            cls.raise_oserr()

    @classmethod
    def lremovexattr(cls, path, attr):
        ret = gr_lremovexattr(cls, path, attr)
        if ret == -1:
            cls.raise_oserr()

    @classmethod
    def llistxattr_buf(cls, path):
        """listxattr variant with size discovery"""
        try:
            # Assuming no more than 100 xattrs in a file/directory and
            # each xattr key length will be less than 256 bytes
            # llistxattr will be called with bigger size so that
            # listxattr will not fail with ERANGE. OSError will be
            # raised if fails even with the large size specified.
            size = 256 * 100
            return cls.llistxattr(path, size)
        except OSError:
            # If fixed length failed for getting list of xattrs then
            # use the llistxattr call to get the size and use that
            # size to get the list of xattrs.
            size = cls.llistxattr(path)
            if size == -1:
                cls.raise_oserr()
            if size == 0:
                return []

            return cls.llistxattr(path, size)



#!/usr/bin/python3

from __future__ import print_function

import os
from errno import EEXIST, ENOENT

from gluster.cliutils import (execute, Cmd, node_output_ok,
                              node_output_notok, execute_in_peers,
                              runcli, oknotok)
from prettytable import PrettyTable

LOG_DIR = "/var/log/glusterfs/geo-replication-secondaries"
CLI_LOG = "/var/log/glusterfs/cli.log"
GEOREP_DIR = "/var/lib/glusterd/geo-replication"
GLUSTERD_VOLFILE = "/etc/glusterfs/glusterd.vol"


class MountbrokerUserMgmt(object):
    def __init__(self, volfile):
        self.volfile = volfile
        self._options = {}
        self.commented_lines = []
        self.user_volumes = {}
        self._parse()

    def _parse(self):
        """ Example glusterd.vol
        volume management
            type mgmt/glusterd
            option working-directory /var/lib/glusterd
            option transport-type socket,rdma
            option transport.socket.keepalive-time 10
            option transport.socket.keepalive-interval 2
            option transport.socket.read-fail-log off
            option rpc-auth-allow-insecure on
            option ping-timeout 0
            option event-threads 1
            # option base-port 49152
            option mountbroker-root /var/mountbroker-root
            option mountbroker-geo-replication.user1 vol1,vol2,vol3
            option geo-replication-log-group geogroup
            option rpc-auth-allow-insecure on
        end-volume
        """
        with open(self.volfile, "r") as f:
            for line in f:
                line = line.strip()
                if line.startswith("option "):
                    key, value = line.split()[1:]
                    self._options[key] = value
                if line.startswith("#"):
                    self.commented_lines.append(line)

        for k, v in self._options.items():
            if k.startswith("mountbroker-geo-replication."):
                user = k.split(".")[-1]
                self.user_volumes[user] = set(v.split(","))

    def get_group(self):
        return self._options.get("geo-replication-log-group", None)

    def _get_write_data(self):
        op = "volume management\n"
        op += "    type mgmt/glusterd\n"
        for k, v in self._options.items():
            if k.startswith("mountbroker-geo-replication."):
                # Users will be added seperately
                continue

            op += "    option %s %s\n" % (k, v)

        for k, v in self.user_volumes.items():
            if v:
                op += ("    option mountbroker-geo-replication."
                       "%s %s\n" % (k, ",".join(v)))

        for line in self.commented_lines:
            op += "    %s\n" % line

        op += "end-volume"
        return op

    def save(self):
        with open(self.volfile + "_tmp", "w") as f:
            f.write(self._get_write_data())
            f.flush()
            os.fsync(f.fileno())
        os.rename(self.volfile + "_tmp", self.volfile)

    def set_mount_root_and_group(self, mnt_root, group):
        self._options["mountbroker-root"] = mnt_root
        self._options["geo-replication-log-group"] = group

    def add(self, volume, user):
        user_volumes = self.user_volumes.get(user, None)

        if user_volumes is not None and volume in user_volumes:
            # User and Volume already exists
            return

        if user_volumes is None:
            # User not exists
            self.user_volumes[user] = set()

        self.user_volumes[user].add(volume)

    def remove(self, volume=None, user=None):
        if user is not None:
            if volume is None:
                self.user_volumes[user] = set()
            else:
                try:
                    self.user_volumes.get(user, set()).remove(volume)
                except KeyError:
                    pass
        else:
            if volume is None:
                return

            for k, v in self.user_volumes.items():
                try:
                    self.user_volumes[k].remove(volume)
                except KeyError:
                    pass

    def info(self):
        # Convert Volumes set into Volumes list
        users = {}
        for k, v in self.user_volumes.items():
            users[k] = list(v)

        data = {
            "mountbroker-root": self._options.get("mountbroker-root", "None"),
            "geo-replication-log-group": self._options.get(
                "geo-replication-log-group", ""),
            "users": users
        }

        return data


class NodeSetup(Cmd):
    # Test if group exists using `getent group <grp>`
    # and then group add using `groupadd <grp>`
    # chgrp -R <grp> /var/log/glusterfs/geo-replication-secondaries
    # chgrp -R <grp> /var/lib/glusterd/geo-replication
    # chmod -R 770 /var/log/glusterfs/geo-replication-secondaries
    # chmod 770 /var/lib/glusterd/geo-replication
    # mkdir -p <mnt_root>
    # chmod 0711 <mnt_root>
    # If selinux,
    # semanage fcontext -a -e /home /var/mountbroker-root
    # restorecon -Rv /var/mountbroker-root
    name = "node-setup"

    def args(self, parser):
        parser.add_argument("mount_root")
        parser.add_argument("group")

    def run(self, args):
        m = MountbrokerUserMgmt(GLUSTERD_VOLFILE)

        try:
            os.makedirs(args.mount_root)
        except OSError as e:
            if e.errno == EEXIST:
                pass
            else:
                node_output_notok("Unable to Create {0}".format(
                    args.mount_root))

        execute(["chmod", "0711", args.mount_root])
        try:
            execute(["semanage", "fcontext", "-a", "-e",
                     "/home", args.mount_root])
        except OSError as e:
            if e.errno == ENOENT:
                pass
            else:
                node_output_notok(
                    "Unable to run semanage: {0}".format(e))

        try:
            execute(["restorecon", "-Rv", args.mount_root])
        except OSError as e:
            if e.errno == ENOENT:
                pass
            else:
                node_output_notok(
                    "Unable to run restorecon: {0}".format(e))

        rc, out, err = execute(["getent", "group", args.group])
        if rc != 0:
            node_output_notok("User Group not exists")

        execute(["chgrp", "-R", args.group, GEOREP_DIR])
        execute(["chgrp", "-R", args.group, LOG_DIR])
        execute(["chgrp", args.group, CLI_LOG])
        execute(["chmod", "770", GEOREP_DIR])
        execute(["find", LOG_DIR, "-type", "d", "-exec", "chmod", "770", "{}",
                 "+"])
        execute(["find", LOG_DIR, "-type", "f", "-exec", "chmod", "660", "{}",
                 "+"])
        execute(["chmod", "660", CLI_LOG])

        m.set_mount_root_and_group(args.mount_root, args.group)
        m.save()

        node_output_ok()


def color_status(value):
    if value.lower() in ("up", "ok", "yes"):
        return "green"
    else:
        return "red"


class CliSetup(Cmd):
    # gluster-mountbroker setup <MOUNT_ROOT> <GROUP>
    name = "setup"

    def args(self, parser):
        parser.add_argument("mount_root",
                            help="Path to the mountbroker-root directory.")
        parser.add_argument("group",
                            help="Group to be used for setup.")

    def run(self, args):
        out = execute_in_peers("node-setup", [args.mount_root,
                                              args.group])
        table = PrettyTable(["NODE", "NODE STATUS", "SETUP STATUS"])
        table.align["NODE STATUS"] = "r"
        table.align["SETUP STATUS"] = "r"
        for p in out:
            table.add_row([p.hostname,
                           "UP" if p.node_up else "DOWN",
                           "OK" if p.ok else "NOT OK: {0}".format(
                               p.error)])

        print(table)


class NodeStatus(Cmd):
    # Check if Group exists
    # Check if user exists
    # Check directory permission /var/log/glusterfs/geo-replication-secondaries
    # and /var/lib/glusterd/geo-replication
    # Check mount root and its permissions
    # Check glusterd.vol file for user, group, dir existance
    name = "node-status"

    def run(self, args):
        m = MountbrokerUserMgmt(GLUSTERD_VOLFILE)
        data = m.info()
        data["group_exists"] = False
        data["path_exists"] = False

        rc, out, err = execute(["getent", "group",
                                data["geo-replication-log-group"]])

        if rc == 0:
            data["group_exists"] = True

        if os.path.exists(data["mountbroker-root"]):
            data["path_exists"] = True

        node_output_ok(data)


class CliStatus(Cmd):
    # gluster-mountbroker status
    name = "status"

    def run(self, args):
        out = execute_in_peers("node-status")
        table = PrettyTable(["NODE", "NODE STATUS", "MOUNT ROOT",
                             "GROUP", "USERS"])
        table.align["NODE STATUS"] = "r"

        for p in out:
            node_data = p.output
            if node_data == "" or node_data == "N/A":
                node_data = {}

            users_row_data = ""
            for k, v in node_data.get("users", {}).items():
                users_row_data += "{0}({1}) ".format(k, ", ".join(v))

            if not users_row_data:
                users_row_data = "None"

            mount_root = node_data.get("mountbroker-root", "None")
            if mount_root != "None":
                mount_root += "({0})".format(oknotok(
                    node_data.get("path_exists", False)))

            grp = node_data.get("geo-replication-log-group", "None")
            if grp != "None":
                grp += "({0})".format(oknotok(
                    node_data.get("group_exists", False)))

            table.add_row([p.hostname,
                           "UP" if p.node_up else "DOWN",
                           mount_root,
                           grp,
                           users_row_data])

        print(table)


class NodeAdd(Cmd):
    # useradd -m -g <grp> <usr>
    # useradd to glusterd.vol
    name = "node-add"

    def args(self, parser):
        parser.add_argument("volume")
        parser.add_argument("user")

    def run(self, args):
        m = MountbrokerUserMgmt(GLUSTERD_VOLFILE)
        grp = m.get_group()
        if grp is None:
            node_output_notok("Group is not available")

        m.add(args.volume, args.user)
        m.save()
        node_output_ok()


class CliAdd(Cmd):
    # gluster-mountbroker add <VOLUME> <USER>
    name = "add"

    def args(self, parser):
        parser.add_argument("volume",
                            help="Volume to be added.")
        parser.add_argument("user",
                            help="User for which volume is to be added.")

    def run(self, args):
        out = execute_in_peers("node-add", [args.volume,
                                            args.user])
        table = PrettyTable(["NODE", "NODE STATUS", "ADD STATUS"])
        table.align["NODE STATUS"] = "r"
        table.align["ADD STATUS"] = "r"

        for p in out:
            table.add_row([p.hostname,
                           "UP" if p.node_up else "DOWN",
                           "OK" if p.ok else "NOT OK: {0}".format(
                               p.error)])

        print(table)


class NodeRemove(Cmd):
    # userremove from glusterd.vol file
    name = "node-remove"

    def args(self, parser):
        parser.add_argument("volume")
        parser.add_argument("user")

    def run(self, args):
        m = MountbrokerUserMgmt(GLUSTERD_VOLFILE)
        volume = None if args.volume == "." else args.volume
        user = None if args.user == "." else args.user
        m.remove(volume=volume, user=user)
        m.save()
        node_output_ok()


class CliRemove(Cmd):
    # gluster-mountbroker remove --volume <VOLUME> --user <USER>
    name = "remove"

    def args(self, parser):
        parser.add_argument("--volume", default=".", help="Volume to be removed.")
        parser.add_argument("--user", default=".",
                            help="User for which volume has to be removed.")

    def run(self, args):
        out = execute_in_peers("node-remove", [args.volume,
                                               args.user])
        table = PrettyTable(["NODE", "NODE STATUS", "REMOVE STATUS"])
        table.align["NODE STATUS"] = "r"
        table.align["REMOVE STATUS"] = "r"

        for p in out:
            table.add_row([p.hostname,
                           "UP" if p.node_up else "DOWN",
                           "OK" if p.ok else "NOT OK: {0}".format(
                               p.error)])

        print(table)

if __name__ == "__main__":
    runcli()
